# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Dummy WorkerPool implementation, for testing."""

from typing import TYPE_CHECKING, override

from django.utils import timezone

from debusine.assets import CloudProvidersType
from debusine.server.worker_pools import WorkerPoolInterface

if TYPE_CHECKING:
    from debusine.db.models import Worker


class DummyWorkerPool(
    WorkerPoolInterface, cloud_provider=CloudProvidersType.DUMMY
):
    """Dummy implementation of WorkerPoolInterface for testing."""

    @override
    def launch_worker(self, worker: "Worker") -> None:
        worker.instance_created_at = timezone.now()
        worker.worker_pool_data = {"launched": True}
        worker.save()

    @override
    def terminate_missing(self) -> None:
        pass



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.
"""Serializers for task inputs."""

import abc
from typing import Any, override

import pydantic

from debusine.tasks.models import (
    ExtraExternalRepository,
    InputArtifactMultiple,
    InputCollectionSingle,
)


class TaskInputSerializer[ValueType](abc.ABC):
    """Serialize/deserialize a TaskInfo value."""

    @abc.abstractmethod
    def serialize(self, value: ValueType) -> Any:
        """Serialize a value to a structure for JSON."""

    @abc.abstractmethod
    def deserialize(self, value: Any) -> ValueType:
        """Deserialize a value from a structure from JSON."""


class ScalarSerializer[ValueType](TaskInputSerializer[ValueType]):
    """Serialize/deserialize a scalar value that already fits in JSON."""

    def __init__(self, value_type: type[ValueType]) -> None:
        """Store the type used to validate serialized values."""
        self.value_type = value_type

    @override
    def serialize(self, value: ValueType) -> Any:
        return value

    @override
    def deserialize(self, value: Any) -> ValueType:
        if not isinstance(value, self.value_type):
            raise ValueError(
                f"Value {value!r} is not an instance of {self.value_type}"
            )
        return value


class PydanticSerializer[ValueType: pydantic.BaseModel](
    TaskInputSerializer[ValueType]
):
    """Serialize/deserialize a pydantic model."""

    def __init__(self, value_type: type[ValueType]) -> None:
        """Store the Pydantic model used to validate serialized values."""
        self.value_type = value_type

    @override
    def serialize(self, value: ValueType) -> Any:
        return value.model_dump(mode="json")

    @override
    def deserialize(self, value: Any) -> ValueType:
        return self.value_type.model_validate(value)


class OptionalPydanticSerializer[ValueType: pydantic.BaseModel](
    TaskInputSerializer[ValueType | None]
):
    """Serialize/deserialize a pydantic model."""

    def __init__(self, value_type: type[ValueType]) -> None:
        """Store the Pydantic model used to validate serialized values."""
        self.value_type = value_type

    @override
    def serialize(self, value: ValueType | None) -> Any:
        if value is None:
            return None
        return value.model_dump(mode="json")

    @override
    def deserialize(self, value: Any) -> ValueType | None:
        if value is None:
            return None
        return self.value_type.model_validate(value)


class PydanticListSerializer[ItemType: pydantic.BaseModel](
    TaskInputSerializer[list[ItemType]]
):
    """Serialize/deserialize a list of pydantic models."""

    def __init__(self, item_type: type[ItemType]) -> None:
        """Store the Pydantic model used to validate serialized values."""
        self.item_type = item_type

    @override
    def serialize(self, value: list[ItemType]) -> Any:
        return [item.model_dump(mode="json") for item in value]

    @override
    def deserialize(self, value: Any) -> list[ItemType]:
        if not isinstance(value, list):
            raise ValueError(f"Value {value!r} is not an instance of {list}")
        return [self.item_type.model_validate(item) for item in value]


class MultiInputListSerializer(
    TaskInputSerializer[list[InputArtifactMultiple]]
):
    """Serialize/deserialize a list of InputArtifactMultiple."""

    @override
    def serialize(self, value: list[InputArtifactMultiple]) -> Any:
        return [item.model_dump(mode="json") for item in value]

    @override
    def deserialize(self, value: Any) -> list[InputArtifactMultiple]:
        if not isinstance(value, list):
            raise ValueError(f"Value {value!r} is not an instance of {list}")

        res: list[InputArtifactMultiple] = [
            InputArtifactMultiple.model_validate(item) for item in value
        ]
        return res


class SuiteArchiveInputSerializer(
    TaskInputSerializer[
        tuple[InputCollectionSingle, InputCollectionSingle | None]
    ]
):
    """Serialize a suite, archive tuple."""

    @override
    def serialize(
        self, value: tuple[InputCollectionSingle, InputCollectionSingle | None]
    ) -> Any:
        return [
            value[0].model_dump(mode="json"),
            (
                value[1].model_dump(mode="json")
                if value[1] is not None
                else None
            ),
        ]

    @override
    def deserialize(
        self, value: Any
    ) -> tuple[InputCollectionSingle, InputCollectionSingle | None]:
        if not isinstance(value, list):
            raise ValueError(f"Value {value!r} is not an instance of {list}")
        if not len(value) == 2:
            raise ValueError(
                f"Value {value!r} should be a list of exactly two items"
            )
        return (
            InputCollectionSingle.model_validate(value[0]),
            (
                InputCollectionSingle.model_validate(value[1])
                if value[1] is not None
                else None
            ),
        )


class ExtraRepositoriesSerializer(
    TaskInputSerializer[list[ExtraExternalRepository] | None]
):
    """Serialize a list of extra repositories."""

    @override
    def serialize(
        self, value: list[ExtraExternalRepository] | None
    ) -> list[dict[str, Any]] | None:
        if value is None:
            return None
        return [item.model_dump(mode="json") for item in value]

    @override
    def deserialize(self, value: Any) -> list[ExtraExternalRepository] | None:
        if value is None:
            return None
        return [ExtraExternalRepository.model_validate(item) for item in value]



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""
View extension to model page descriptions.

A Place describes information about a page that can be used to generate
elements like page titles and breadcrumb elements.
"""

import abc
from enum import StrEnum
from typing import Any, NotRequired, TypedDict, Unpack, override

from django.utils.html import format_html
from django.utils.safestring import SafeString


class PlaceType(StrEnum):
    """Supported place types."""

    INDEX = "index"
    RESOURCE = "resource"
    CONTAINER = "container"


class PlaceKwargs(TypedDict):
    """Argument typing for Place constructor."""

    #: Title
    title: str
    #: Description
    description: NotRequired[str | None]
    #: URL
    url: str
    #: Icon
    icon: NotRequired[str | None]
    #: Parent place (used for breadcrumbs)
    parent: NotRequired["BreadcrumbPlace | None"]


class Place(abc.ABC):
    """Describe a page in Debusine's web space."""

    def __init__(self, **kwargs: Unpack[PlaceKwargs]) -> None:
        """Store Place attributes."""
        #: Title
        self.title = kwargs["title"]
        #: Description
        self.description = kwargs.get("description")
        #: URL
        self.url = kwargs["url"]
        #: Icon
        self.icon = kwargs.get("icon")
        #: Parent place (used for breadcrumbs)
        self.parent = kwargs.get("parent")

    @abc.abstractmethod
    def get_type(self) -> PlaceType:
        """Get the view type for this place."""

    @abc.abstractmethod
    def is_page_title_visible(self) -> bool:
        """Return True if the page title should be shown."""

    def as_icon(self, **kwargs: Any) -> str:
        """
        Render the icon.

        :returns: the rendered HTML, or an empty string if this Place has no
          icon
        """
        if self.icon is None:
            return SafeString()
        else:
            return format_html(
                """<span class="bi bi-{icon}"></span> """, icon=self.icon
            )

    def as_tooltip(self, **kwargs: Any) -> str:
        """Render as a title= attribute."""
        if self.description is not None:
            return format_html(
                ' title="{description}"', description=self.description
            )
        else:
            return SafeString()

    def as_nav(self, **kwargs: Any) -> str:
        """Render as a <nav> link."""
        return format_html(
            """<a class="nav-link btn btn-debusine" href="{url}"{desc}>"""
            "{icon}{title}</a>",
            url=self.url,
            icon=self.as_icon(),
            title=self.title,
            desc=self.as_tooltip(),
        )

    def as_head_title(self, **kwargs: Any) -> str:
        """Render as a <title> element."""
        if self.parent is None:
            prefix = ""
        else:
            prefix = " - ".join(p.breadcrumb for p in self.parent.parents())
            prefix += " - "
        return format_html(
            "<title>{prefix}{title}</title>", prefix=prefix, title=self.title
        )

    @abc.abstractmethod
    def as_page_title(self, **kwargs: Any) -> str:
        """Render as a page title."""

    @abc.abstractmethod
    def as_breadcrumbs(self, **kwargs: Any) -> str:
        """Render as breadcrumbs leading to this Place."""

    def as_button(self, **kwargs: Any) -> str:
        """
        Render as a button.

        Supported keyword arguments:

        * ``label``: override button label
        """
        # TODO: support a way for templates to pass class information
        return format_html(
            """<a class="btn btn-primary" href="{url}"{tooltip}>"""
            """{icon}{title}</a>""",
            url=self.url,
            title=kwargs.get("label", self.title),
            icon=self.as_icon(),
            tooltip=self.as_tooltip(),
        )


class PlaceWithoutPageTitle(Place):
    """Place which does not render a page title."""

    @override
    def is_page_title_visible(self) -> bool:
        return False

    @override
    def as_page_title(self, **kwargs: Any) -> str:
        return SafeString()


class PlaceWithPageTitle(Place):
    """Place which renders a page title."""

    @override
    def is_page_title_visible(self) -> bool:
        return True

    @override
    def as_page_title(self, **kwargs: Any) -> str:
        return format_html(
            """<h1 class="mb-4">{title}</h1>""", title=self.title
        )


class IndexPlace(PlaceWithoutPageTitle):
    """Place showing an overview or a list of resources."""

    @override
    def get_type(self) -> PlaceType:
        return PlaceType.INDEX

    @override
    def as_breadcrumbs(self, **kwargs: Any) -> str:
        if self.parent:
            parent = self.parent.as_breadcrumb_parent()
        else:
            parent = SafeString()
        return format_html(
            """{parent}<h1 class="breadcrumb-current">{title}</h1>""",
            parent=parent,
            title=self.title,
        )


class BreadcrumbPlaceKwargs(PlaceKwargs):
    """Argument typing for BreadcrumbPlace constructor."""

    #: Short label used for breadcrumbs
    breadcrumb: str


class BreadcrumbPlace(Place, abc.ABC):
    """Place supporting showing in the breadcrumbs bar."""

    def __init__(
        self, *, breadcrumb: str, **kwargs: Unpack[PlaceKwargs]
    ) -> None:
        """Store Place attributes."""
        super().__init__(**kwargs)
        #: Short label used for breadcrumbs
        self.breadcrumb = breadcrumb

    def parents(self, **kwargs: Any) -> list["BreadcrumbPlace"]:
        """Build the parent chain as a list, topmost parent first, self last."""
        if self.parent is None:
            res = []
        else:
            res = self.parent.parents()
        res.append(self)
        return res

    def as_breadcrumb_parent(self, **kwargs: Any) -> str:
        """Render as a parent in a breadcrumb list."""
        if self.parent:
            parent = self.parent.as_breadcrumb_parent()
        else:
            parent = SafeString()
        return format_html(
            """{parent}<a href="{url}" class="breadcrumb-parent"{tooltip}>"""
            """{breadcrumb}</a>"""
            """<span class="breadcrumb-separator"></span>""",
            parent=parent,
            url=self.url,
            tooltip=self.as_tooltip(),
            breadcrumb=self.breadcrumb,
        )


class ResourcePlaceKwargs(BreadcrumbPlaceKwargs):
    """Argument typing for ResourcePlace constructor."""


class ResourcePlace(BreadcrumbPlace, PlaceWithPageTitle):
    """Place showing details about a resource."""

    @override
    def get_type(self) -> PlaceType:
        return PlaceType.RESOURCE

    @override
    def as_breadcrumbs(self, **kwargs: Any) -> str:
        if self.parent:
            parent = self.parent.as_breadcrumb_parent()
        else:
            parent = SafeString()
        return format_html(
            "{parent}"
            """<span class="breadcrumb-current"{tooltip}>{title}</span>""",
            parent=parent,
            tooltip=self.as_tooltip(),
            title=self.breadcrumb,
        )


class ContainerPlaceKwargs(BreadcrumbPlaceKwargs):
    """Argument typing for ContainerPlace constructor."""


class ContainerPlace(BreadcrumbPlace, PlaceWithoutPageTitle):
    """Place showing a group of resources."""

    @override
    def get_type(self) -> PlaceType:
        return PlaceType.CONTAINER

    @override
    def as_breadcrumbs(self, **kwargs: Any) -> str:
        if self.parent:
            parent = self.parent.as_breadcrumb_parent()
        else:
            parent = SafeString()
        return format_html(
            "{parent}"
            """<span class="breadcrumb-current"{tooltip}>{title}</span>""",
            parent=parent,
            tooltip=self.as_tooltip(),
            title=self.title,
        )



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Tests for table columns."""

from typing import override
from unittest import mock

from django.test import RequestFactory

from debusine.db.models import User
from debusine.test.django import TestCase
from debusine.web.views.table import (
    Column,
    DateTimeColumn,
    NumberColumn,
    Ordering,
    Sort,
    StringColumn,
    Table,
)
from debusine.web.views.table.columns import BoundColumn, Columns


class BoundColumnTests(TestCase):
    """Tests for :py:class:`BoundColumn`."""

    def _column(self, column: Column) -> BoundColumn[User]:
        """Create a column bound to a mock table."""

        class _Table(Table[User]):
            @override
            def init(self) -> None:
                super().init()
                self.add_column(column.name, column)

        table = _Table(
            RequestFactory().get("/"),
            User.objects.all(),
            default_order=column.name,
        )

        return table.columns[column.name]

    def assertOrderings(self, col: BoundColumn[User], names: list[str]) -> None:
        """Ensure the orderings match the given "what.how" names."""
        self.assertEqual([f"{o.what}.{o.how}" for o in col.orderings], names)

    def test_column(self) -> None:
        """Test Column defaulting to StringColumn."""
        col = self._column(Column("User", ordering="username", name="user"))
        self.assertOrderings(col, ["user.asc", "user.desc"])
        self.assertEqual(str(col), "user")
        self.assertEqual(repr(col), "BoundColumn('user')")
        self.assertEqual(col.name, "user")
        self.assertEqual(col.title, "User")
        ctx = col.get_context_data()
        self.assertEqual(ctx["column"], col)
        self.assertHTMLValid(col.sort_icon)

    def test_string_column(self) -> None:
        """Test StringColumn."""
        col = self._column(
            StringColumn("Email", ordering="email", name="email")
        )
        self.assertOrderings(col, ["email.asc", "email.desc"])
        self.assertEqual(str(col), "email")
        self.assertEqual(repr(col), "BoundColumn('email')")
        self.assertEqual(col.name, "email")
        self.assertEqual(col.title, "Email")
        ctx = col.get_context_data()
        self.assertEqual(ctx["column"], col)
        self.assertHTMLValid(col.sort_icon)

    def test_number_column(self) -> None:
        """Test NumberColumn."""
        col = self._column(NumberColumn("Age", ordering="age", name="age"))
        self.assertOrderings(col, ["age.asc", "age.desc"])
        self.assertEqual(str(col), "age")
        self.assertEqual(repr(col), "BoundColumn('age')")
        self.assertEqual(col.name, "age")
        self.assertEqual(col.title, "Age")
        self.assertEqual(col.get_context_data(), {"column": col, "widget": col})
        self.assertHTMLValid(col.sort_icon)

    def test_datetime_column(self) -> None:
        """Test DateTimeColumn."""
        col = self._column(
            DateTimeColumn("Last Seen", ordering="last_seen", name="last_seen")
        )
        self.assertOrderings(col, ["last_seen.asc", "last_seen.desc"])
        self.assertEqual(str(col), "last_seen")
        self.assertEqual(repr(col), "BoundColumn('last_seen')")
        self.assertEqual(col.name, "last_seen")
        self.assertEqual(col.title, "Last Seen")
        self.assertEqual(col.get_context_data(), {"column": col, "widget": col})
        self.assertHTMLValid(col.sort_icon)

    def test_custom_sort(self) -> None:
        """Test DateTimeColumn."""
        col = self._column(
            Column(
                "Test",
                ordering=Ordering(foo=Sort("username", "Foo", "icon")),
                name="test",
            )
        )
        self.assertOrderings(col, ["test.foo"])
        self.assertEqual(str(col), "test")
        self.assertEqual(repr(col), "BoundColumn('test')")
        self.assertEqual(col.name, "test")
        self.assertEqual(col.title, "Test")
        self.assertEqual(col.get_context_data(), {"column": col, "widget": col})
        self.assertHTMLValid(col.sort_icon)

    def test_unsortable_column(self) -> None:
        """Test data of unsortable columns."""
        col = self._column(Column("Extra", name="extra"))
        self.assertEqual(col.orderings, [])
        self.assertEqual(str(col), "extra")
        self.assertEqual(repr(col), "BoundColumn('extra')")
        self.assertEqual(col.name, "extra")
        self.assertEqual(col.title, "Extra")
        self.assertEqual(col.get_context_data(), {"column": col, "widget": col})
        self.assertEqual(col.sort_icon, "")


class ColumnsTests(TestCase):
    """Tests for :py:class:`Columns`."""

    def _columns(self) -> Columns[User]:
        """Create a Columns bound to a mock table."""
        table = mock.Mock()
        table.request = RequestFactory().get("/")
        return Columns(table)

    def test_init(self) -> None:
        columns = self._columns()
        self.assertEqual(columns.columns, {})

    def test_add(self) -> None:
        columns = self._columns()
        col = columns.add(
            "user", Column("User", ordering="username", name="user")
        )
        self.assertIs(col.table, columns.table)
        self.assertEqual(columns.columns, {"user": col})

    def test_getitem(self) -> None:
        columns = self._columns()
        col = columns.add(
            "user", Column("User", ordering="username", name="user")
        )
        self.assertIs(columns["user"], col)
        with self.assertRaises(KeyError):
            columns["foo"]



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Concrete but mock version of FieldResolver used for tests."""

from collections.abc import Collection as AbcCollection
from typing import Any, overload, override

from debusine.artifacts.models import (
    ArtifactCategory,
    ArtifactData,
    CollectionCategory,
)
from debusine.client.models import RelationType
from debusine.tasks.executors import ExecutorImageCategory
from debusine.tasks.inputs import inputs
from debusine.tasks.inputs.resolver import FieldResolver
from debusine.tasks.models import (
    BackendType,
    BaseTaskInputArtifact,
    ExtraExternalRepository,
    InputArtifactMultiple,
    InputArtifactSingle,
    InputCollectionSingle,
    LookupMultiple,
    LookupSingle,
)


class ConcreteFieldResolver(FieldResolver):
    """
    Concrete FieldResolver implementation.

    All abstract methods are overridden with versions that raise
    NotImplementedError.

    This is useful as a base for mocking specific methods.
    """

    @overload
    def lookup_single_artifact(
        self,
        lookup: LookupSingle,
        *,
        default_category: CollectionCategory | None = None,
        label: str,
    ) -> InputArtifactSingle: ...

    @overload
    def lookup_single_artifact(
        self,
        lookup: None,
        *,
        default_category: CollectionCategory | None = None,
        label: str,
    ) -> None: ...

    @override
    def lookup_single_artifact(
        self,
        lookup: LookupSingle | None,
        *,
        default_category: CollectionCategory | None = None,
        label: str,
    ) -> InputArtifactSingle | None:
        raise NotImplementedError("lookup_single_artifact")

    @override
    def lookup_multiple_artifacts(
        self,
        lookup: LookupMultiple,
        *,
        default_category: CollectionCategory | None = None,
        label: str,
    ) -> InputArtifactMultiple:
        raise NotImplementedError("lookup_multiple_artifact")

    @override
    def find_related_artifacts(
        self,
        artifact_ids: AbcCollection[int],
        *,
        target_category: ArtifactCategory,
        relation_type: RelationType = RelationType.RELATES_TO,
        label: str,
    ) -> InputArtifactMultiple:
        raise NotImplementedError("find_related_artifacts")

    @override
    def get_server_setting(self, setting: str) -> str:
        raise NotImplementedError("get_server_setting")

    @override
    def get_fallback_architecture(self) -> str:
        raise NotImplementedError("get_fallback_architecture")

    @override
    def export_suite_signing_keys(self, suite_id: int) -> str:
        raise NotImplementedError("export_suite_signing_keys")

    @override
    def lookup_environment(
        self,
        lookup: LookupSingle,
        *,
        architecture: str | None = None,
        backend: BackendType | None = None,
        default_category: CollectionCategory | None = None,
        image_category: "ExecutorImageCategory | None" = None,
        try_variant: str | None = None,
        label: str,
    ) -> InputArtifactSingle:
        raise NotImplementedError("lookup_environment")

    @override
    def resolve_environmentinput(
        self, field: inputs.EnvironmentInput
    ) -> InputArtifactSingle:
        raise NotImplementedError("resolve_environmentinput")

    @override
    def resolve_optionalenvironmentinput(
        self, field: inputs.OptionalEnvironmentInput
    ) -> InputArtifactSingle | None:
        raise NotImplementedError("resolve_optionalenvironmentinput")

    @override
    def resolve_suitearchiveinput(
        self, field: inputs.SuiteArchiveInput
    ) -> tuple[InputCollectionSingle, InputCollectionSingle | None]:
        raise NotImplementedError("resolve_suitearchiveinput")

    @override
    def resolve_extrarepositoriesinput(
        self, field: inputs.ExtraRepositoriesInput
    ) -> list[ExtraExternalRepository] | None:
        raise NotImplementedError("resolve_extrarepositoriesinput")

    @override
    def get_field[T](self, field: "inputs.DataFieldInput[T, Any]") -> T | None:
        raise NotImplementedError("get_field")


class MockFieldResolver(ConcreteFieldResolver):
    """FieldResolver that fill inputs with a set of known values."""

    def __init__(self, **field_values: Any) -> None:
        """Construct a :py:class:`MockTaskDatabase`."""
        self.field_values: dict[str, Any] = {
            "build_architecture": "amd64",
            "task_configuration": None,
            **field_values,
        }

    def set(self, **field_values: Any) -> None:
        """Set field values from keyword arguments."""
        self.field_values.update(field_values)

    def set_field_artifact(
        self,
        name: str,
        *,
        id: int,  # noqa: A002
        category: ArtifactCategory,
        data: ArtifactData,
    ) -> None:
        """Set the value of an input field for an artifact lookup."""
        self.field_values[name] = InputArtifactSingle(
            artifact_id=id, category=category, data=data, label=name, lookup=id
        )

    def set_field_artifacts(
        self,
        name: str,
        *artifacts: BaseTaskInputArtifact,
    ) -> None:
        """Set the value of an input field for a multiple artifact lookup."""
        self.field_values[name] = InputArtifactMultiple(
            label=name, lookup=["test"], artifacts=artifacts
        )

    def set_field_artifact_list(
        self,
        name: str,
        *artifacts: BaseTaskInputArtifact,
    ) -> None:
        """Set the value for a list of single artifact lookups."""
        self.field_values[name] = list(artifacts)

    def set_field_artifact_difflist(
        self,
        name: str,
        *lists: list[BaseTaskInputArtifact],
    ) -> None:
        """Set a list of two multiple lookups for debdiff."""
        self.field_values[name] = [
            InputArtifactMultiple(label=name, lookup=["test"], artifacts=lst)
            for lst in lists
        ]

    @override
    def resolve_inputs(
        self, task: inputs.TaskFieldContainer, stage: inputs.Stage | None = None
    ) -> None:
        for cont, inp in task.iter_all_fields(stage):
            assert inp.name in self.field_values, (
                f"Value for field {cont.__class__.__name__}.{inp.name}"
                " was not provided"
            )
            cont.__dict__[inp.name] = self.field_values[inp.name]



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Models used for asset data."""

import abc
from enum import StrEnum
from typing import Annotated, Any, Literal, override

import pydantic

from debusine.utils import Base64BytesAsStr, DjangoChoicesEnum


class AssetCategory(DjangoChoicesEnum):
    """Possible asset categories."""

    CLOUD_PROVIDER_ACCOUNT = "debusine:cloud-provider-account"
    SIGNING_KEY = "debusine:signing-key"
    APT_AUTHENTICATION = "debian:apt-authentication"


class BaseAssetDataModel(pydantic.BaseModel, abc.ABC):
    """Base pydantic model for asset data and their components."""

    model_config = pydantic.ConfigDict(validate_assignment=True, extra="forbid")


class ModelCategory(BaseAssetDataModel):
    """All models for a category, organized by discriminator value."""

    discriminator: str | None = None
    models: dict[str | None, type[BaseAssetDataModel]] = {}


_data_models: dict[AssetCategory, ModelCategory] = {}


class BaseRegisteredAssetDataModel(BaseAssetDataModel, abc.ABC):
    """An asset data model registered for use by `asset_data_model`."""

    @override
    def __init_subclass__(
        cls,
        category: AssetCategory,
        discriminator: str | None = None,
        **kwargs: Any,
    ) -> None:
        """Register subclass in _data_models."""
        super().__init_subclass__(**kwargs)
        if category not in _data_models:
            _data_models[category] = ModelCategory(discriminator=discriminator)
        modelcategory = _data_models[category]
        assert modelcategory.discriminator == discriminator
        discriminator_value: str | None = None
        if discriminator:
            discriminator_value = getattr(cls, discriminator)
        modelcategory.models[discriminator_value] = cls


def asset_data_model(
    category: AssetCategory | str,
    data: dict[str, Any],
) -> BaseAssetDataModel:
    """Load the DataModel for an asset, with specified category and data."""
    category = AssetCategory(category)
    if category not in _data_models:
        raise ValueError(f"No data model for {category} exists.")
    modelcategory = _data_models[category]
    if modelcategory.discriminator is None:
        model = modelcategory.models[None]
    else:
        if modelcategory.discriminator not in data:
            raise ValueError(
                f"Discriminator key {modelcategory.discriminator} "
                f"required for {category} missing in data."
            )
        discriminator_value = data[modelcategory.discriminator]
        if discriminator_value not in modelcategory.models:
            raise ValueError(
                f"No data model for {category} with "
                f"{modelcategory.discriminator}={discriminator_value}."
            )
        model = modelcategory.models[discriminator_value]
    return model(**data)


def asset_categories() -> list[AssetCategory]:
    """List all known asset categories."""
    return list(_data_models.keys())


class KeyPurpose(DjangoChoicesEnum):
    """Choices for SigningKey.purpose."""

    # Sign a UEFI Secure Boot image.
    UEFI = "uefi"

    # Sign a Debian upload (.changes) using OpenPGP.  (The value should
    # perhaps be "openpgp-upload", but is kept this way for compatibility.)
    OPENPGP_UPLOAD = "openpgp"

    # Sign Debian repository indexes using OpenPGP.
    OPENPGP_REPOSITORY = "openpgp-repository"


class SigningKeyData(
    BaseRegisteredAssetDataModel, category=AssetCategory.SIGNING_KEY
):
    """Data for a debusine:signing-key asset."""

    purpose: KeyPurpose
    fingerprint: str
    public_key: Base64BytesAsStr
    description: str | None


class CloudProvidersType(StrEnum):
    """Choices for CloudProviderAccountData.provider_type."""

    AWS = "aws"
    HETZNER = "hetzner"
    DUMMY = "dummy"


class AWSProviderAccountConfiguration(BaseAssetDataModel):
    """Non-secret configuration for an AWS account."""

    region_name: str | None = None
    ec2_endpoint_url: str | None = None
    s3_endpoint_url: str | None = None


class AWSProviderAccountCredentials(BaseAssetDataModel):
    """Secret credentials for an AWS account."""

    access_key_id: str
    secret_access_key: str


class AWSProviderAccountData(
    BaseRegisteredAssetDataModel,
    category=AssetCategory.CLOUD_PROVIDER_ACCOUNT,
    discriminator="provider_type",
):
    """Data for a debusine:cloud-provider-account asset for AWS."""

    provider_type: Literal[CloudProvidersType.AWS] = CloudProvidersType.AWS
    name: str
    configuration: AWSProviderAccountConfiguration
    credentials: AWSProviderAccountCredentials


class HetznerProviderAccountConfiguration(BaseAssetDataModel):
    """Non-secret configuration for a Hetzner Cloud account."""

    region_name: str


class HetznerProviderAccountCredentials(BaseAssetDataModel):
    """Secret credentials for a Hetzner Cloud account."""

    api_token: str


class HetznerProviderAccountData(
    BaseRegisteredAssetDataModel,
    category=AssetCategory.CLOUD_PROVIDER_ACCOUNT,
    discriminator="provider_type",
):
    """Data for a debusine:cloud-provider-account asset for Hetzner Cloud."""

    provider_type: Literal[CloudProvidersType.HETZNER] = (
        CloudProvidersType.HETZNER
    )
    name: str
    configuration: HetznerProviderAccountConfiguration
    credentials: HetznerProviderAccountCredentials


class DummyProviderAccountData(
    BaseRegisteredAssetDataModel,
    category=AssetCategory.CLOUD_PROVIDER_ACCOUNT,
    discriminator="provider_type",
):
    """Data for a debusine:cloud-provider-account asset for test cases."""

    provider_type: Literal[CloudProvidersType.DUMMY] = CloudProvidersType.DUMMY
    name: str
    secret: str = "secret"


CloudProviderAccountData = Annotated[
    AWSProviderAccountData
    | HetznerProviderAccountData
    | DummyProviderAccountData,
    pydantic.Field(discriminator="provider_type"),
]


class APTAuthenticationScheme(StrEnum):
    """Choices for APTAuthentication.auth_scheme."""

    BASIC = "basic"


class BasicAPTCredentials(BaseAssetDataModel):
    """APT credentials for use with HTTP Basic Auth."""

    username: str
    password: str


class BasicAPTAuthenticationData(
    BaseRegisteredAssetDataModel,
    category=AssetCategory.APT_AUTHENTICATION,
    discriminator="auth_scheme",
):
    """Data for a debian:apt-authentication asset for HTTP Basic Auth."""

    auth_scheme: Literal[APTAuthenticationScheme.BASIC] = (
        APTAuthenticationScheme.BASIC
    )
    name: str
    credentials: BasicAPTCredentials


APTAuthenticationData = Annotated[
    BasicAPTAuthenticationData, pydantic.Field(discriminator="auth_scheme")
]



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Tests for the workflow views."""

import math
from typing import ClassVar, override
from unittest.mock import patch

import lxml
from django.urls import reverse
from django.utils import timezone
from rest_framework import status

from debusine.db.models import WorkRequest, WorkflowTemplate
from debusine.db.playground import scenarios
from debusine.test.django import TestCase
from debusine.web.views.tests.utils import ViewTestMixin
from debusine.web.views.workflows import WorkflowsListView


class WorkflowsListViewTests(ViewTestMixin, TestCase):
    """Tests for WorkflowsListView class."""

    template: ClassVar[WorkflowTemplate]
    workflow_1: ClassVar[WorkRequest]
    workflow_2: ClassVar[WorkRequest]

    scenario = scenarios.DefaultContext(set_current=True)

    @override
    @classmethod
    def setUpTestData(cls) -> None:
        super().setUpTestData()

        cls.template = cls.playground.create_workflow_template(
            "name-2", "noop", workspace=cls.scenario.workspace
        )

        cls.workflow_1 = cls.playground.create_workflow(
            cls.template,
            task_data={},
        )
        cls.workflow_2 = cls.playground.create_workflow(
            cls.template,
            task_data={},
        )

    def assertWorkflowRow(
        self,
        row: lxml.objectify.ObjectifiedElement,
        workflow: WorkRequest,
        expected_status: str,
        expected_success: int = 0,
        expected_failure: int = 0,
        expected_pending: int = 0,
        expected_blocked: int = 0,
    ) -> None:
        """Ensure the row shows the given work request."""
        line1 = row.div[0]
        assert workflow.workflow_display_name_parameters
        self.assertTextContentEqual(
            line1.div[0], workflow.workflow_display_name_parameters
        )
        self.assertTextContentEqual(line1.div[1], expected_status)
        self.assertTextContentEqual(
            line1.div[2],
            f"{expected_success}-{expected_failure}"
            f" {expected_pending}-{expected_blocked}",
        )

        line2 = row.div[1]
        self.assertTextContentEqual(line2.div.a, f"#{workflow.id}")
        self.assertEqual(line2.div.a.get("href"), workflow.get_absolute_url())

    def test_can_display_used(self) -> None:
        """Permissions are used via can_display()."""
        self.client.force_login(self.scenario.user)

        with patch(
            "debusine.db.models.work_requests.WorkRequestQuerySet.can_display"
        ) as mock_can_display:
            mock_can_display.return_value = WorkRequest.objects.all()

            self.client.get(
                reverse(
                    "workspaces:workflows:list",
                    kwargs={"wname": self.scenario.workspace.name},
                ),
            )

        mock_can_display.assert_called_once()

    def test_no_workflows(self) -> None:
        """View details shows "No workflows" if no workflows in the space."""
        WorkRequest.objects.all().delete()

        response = self.client.get(
            reverse(
                "workspaces:workflows:list",
                kwargs={"wname": self.scenario.workspace.name},
            ),
        )

        self.assertContains(response, "<p>No workflows.</p>", html=True)

    def test_list_no_filtering_check_workspace(self) -> None:
        """View detail return all workflows for the specific workspace."""
        workspace_unused = self.playground.create_workspace(
            name="unused-workspace", public=True
        )
        # The following workflow is not in the workspace that will be used
        template_unused = self.playground.create_workflow_template(
            "name-1", "noop", workspace=workspace_unused
        )
        self.playground.create_workflow(template_unused, task_data={})

        wr_success = self.workflow_1.create_child_worker("noop")
        self.playground.advance_work_request(
            wr_success, result=WorkRequest.Results.SUCCESS
        )
        wr_pending = self.workflow_1.create_child_worker("noop")
        self.playground.advance_work_request(wr_pending, mark_pending=True)

        self.workflow_1.started_at = timezone.now()
        self.workflow_1.save()

        self.workflow_2.mark_completed(result=WorkRequest.Results.SUCCESS)

        response = self.client.get(
            reverse(
                "workspaces:workflows:list",
                kwargs={"wname": self.scenario.workspace.name},
            ),
        )
        tree = self.assertResponseHTML(response)
        link = (
            "https://freexian-team.pages.debian.net/"
            "debusine/explanation/concepts.html#workflows"
        )
        self.assertContains(
            response,
            f'<p>This page lets you monitor '
            f'<a href="{link}">workflows</a> started in the '
            f'{self.scenario.workspace} workspace.</p>',
            html=True,
        )

        table = self.assertHasElement(tree, "//div[@id='workflow-list-table']")
        rows = self.assertHasElement(
            table, "div[@id='workflow-list-table-rows']"
        )

        # One work request is in a different workspace
        self.assertEqual(len(rows.div), 2)

        self.assertWorkflowRow(
            rows.div[0], self.workflow_2, expected_status="C S"
        )
        self.assertWorkflowRow(
            rows.div[1],
            self.workflow_1,
            expected_status="P",
            expected_success=1,
            expected_pending=1,
        )

    def test_status_beyond_last_page(self) -> None:
        """200, even if page number exceeds number of pages"""
        num_objects = WorkRequest.objects.count()
        paginate_by = WorkflowsListView.paginate_by
        num_pages = math.ceil(num_objects / paginate_by)
        page_num = num_pages + 1

        response = self.client.get(
            reverse(
                "workspaces:workflows:list",
                kwargs={
                    "wname": self.scenario.workspace.name,
                },
            ),
            {"page": page_num},
        )
        self.assertEqual(response.status_code, status.HTTP_200_OK)



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Tests for the management command workspace."""

import datetime as dt
import io
from typing import ClassVar, override

import yaml
from django.core.management import CommandError

from debusine.artifacts.models import (
    ArtifactCategory,
    CollectionCategory,
    SINGLETON_COLLECTION_CATEGORIES,
)
from debusine.db.models import (
    ArtifactRelation,
    Scope,
    WorkflowTemplate,
    Workspace,
    default_workspace,
)
from debusine.db.models.workspaces import WorkspaceRole
from debusine.db.playground import Playground, scenarios
from debusine.django.management.tests import call_command
from debusine.server.management.commands.tests.utils import TabularOutputTests
from debusine.test.django import TestCase


class WorkspaceCommandsTests(TabularOutputTests, TestCase):
    """Tests for workspace management commands."""

    scope: ClassVar[Scope]

    @override
    @classmethod
    def setUpTestData(cls) -> None:
        super().setUpTestData()
        cls.scope = cls.playground.get_or_create_scope(name="localtest")

    def test_define_scenario(self) -> None:
        """Test define command."""
        # define/create, with errors
        with self.assertRaisesRegex(
            CommandError,
            "Error: owners_group is required when creating workspace",
        ):
            call_command("workspace", "create", "localtest/myws")

        with self.assertRaisesRegex(
            CommandError,
            "scope_workspace 'myws' should be in the form"
            " 'scopename/workspacename'",
        ):
            call_command("workspace", "create", "myws", "Owners-myws")

        with self.assertRaisesRegex(
            CommandError,
            "Scope 'nonexistent' not found",
        ):
            call_command(
                "workspace", "create", "nonexistent/my ws", "Owners-myws"
            )

        with self.assertRaisesRegex(
            CommandError,
            "Error creating workspace: 'my ws' is not a valid workspace name",
        ):
            call_command(
                "workspace", "create", "localtest/my ws", "Owners-myws"
            )

        with self.assertRaisesRegex(
            CommandError, "Group 'Owners-myws' not found in scope 'localtest'"
        ):
            call_command("workspace", "create", "localtest/myws", "Owners-myws")

        self.playground.create_group("Owners-myws", scope=self.scope)
        self.playground.create_group("Owners-myws2", scope=self.scope)

        # create/define, idempotent
        for _ in (1, 2):
            stdout, stderr, exit_code = call_command(
                "workspace", "define", "localtest/myws", "Owners-myws"
            )
            # assert workspace and role
            self.assertEqual(stdout, "")
            self.assertEqual(stderr, "")
            self.assertEqual(exit_code, 0)
            workspace = Workspace.objects.get(scope=self.scope, name="myws")
            role = WorkspaceRole.objects.get(
                resource=workspace, group__name="Owners-myws"
            )
            self.assertEqual(role.role, WorkspaceRole.Roles.OWNER)
            self.assertFalse(workspace.public)
            self.assertEqual(
                workspace.default_expiration_delay, dt.timedelta(30)
            )
            self.assertQuerySetEqual(
                workspace.collections.values_list("category", flat=True),
                SINGLETON_COLLECTION_CATEGORIES,
                ordered=False,
            )

        # Set parameters
        call_command(
            "workspace",
            "define",
            "localtest/myws-new",
            "Owners-myws",
            "--public",
            "--default-expiration-delay",
            "0",
            "--no-singleton-collections",
        )
        workspace = Workspace.objects.get(scope=self.scope, name="myws-new")
        self.assertTrue(workspace.public)
        self.assertEqual(workspace.default_expiration_delay, dt.timedelta(0))
        self.assertQuerySetEqual(
            workspace.collections.values_list("category", flat=True), []
        )

        # Change parameters
        call_command(
            "workspace",
            "define",
            "localtest/myws-new",
            "Owners-myws2",
            "--private",
            "--default-expiration-delay",
            "10",
        )
        workspace.refresh_from_db()
        self.assertFalse(workspace.public)
        self.assertEqual(workspace.default_expiration_delay, dt.timedelta(10))
        role = WorkspaceRole.objects.get(
            resource=workspace, group__name="Owners-myws2"
        )
        self.assertEqual(role.role, WorkspaceRole.Roles.OWNER)

        # Don't reset parameters to defaults
        call_command(
            "workspace",
            "define",
            "localtest/myws-new",
        )
        self.assertFalse(workspace.public)
        self.assertEqual(workspace.default_expiration_delay, dt.timedelta(10))

    def test_rename_scenario(self) -> None:
        """Test rename command."""
        self.playground.create_group("Owners-myws", scope=self.scope)
        call_command("workspace", "create", "localtest/myws", "Owners-myws")

        # rename, with errors
        with self.assertRaisesRegex(
            CommandError, "Workspace 'toto' not found in scope 'localtest'"
        ):
            call_command("workspace", "rename", "localtest/toto", "toto2")

        _, stderr, exit_code = call_command(
            "workspace", "rename", "localtest/myws", "my ws"
        )
        self.assertEqual(
            stderr,
            """Renamed workspace would be invalid:
* name: 'my ws' is not a valid workspace name
""",
        )
        self.assertEqual(exit_code, 3)
        self.assertTrue(
            Workspace.objects.filter(scope=self.scope, name="myws").exists()
        )

        call_command(
            "workspace", "create", "localtest/myws-exists", "Owners-myws"
        )
        _, stderr, exit_code = call_command(
            "workspace", "rename", "localtest/myws", "myws-exists"
        )
        self.assertEqual(
            stderr,
            """Renamed workspace would be invalid:
* __all__: Workspace with this Scope and Name already exists.
""",
        )
        self.assertEqual(exit_code, 3)

        # rename, idempotent
        _, stderr, exit_code = call_command(
            "workspace", "rename", "localtest/myws", "myws-renamed"
        )
        self.assertEqual(exit_code, 0)
        self.assertFalse(
            Workspace.objects.filter(scope=self.scope, name="myws").exists()
        )
        self.assertTrue(
            Workspace.objects.filter(
                scope=self.scope, name="myws-renamed"
            ).exists()
        )

        _, stderr, exit_code = call_command(
            "workspace", "rename", "localtest/myws-renamed", "myws-renamed"
        )
        self.assertEqual(exit_code, 0)

    def test_list(self) -> None:
        """Test list command."""
        # Setup
        self.playground.create_group("Owners-myws", scope=self.scope)

        call_command(
            "workspace", "create", "localtest/public", "Owners-myws", "--public"
        )
        call_command("workspace", "create", "localtest/otherfs", "Owners-myws")
        call_command(
            "workspace",
            "create",
            "localtest/noexpire",
            "Owners-myws",
            "--default-expiration-delay",
            "0",
        )

        # list
        with self.assertPrintsTable() as output:
            stdout, _, _ = call_command("workspace", "list")

        self.assertEqual(
            output.col(0),
            [
                "debusine/System",
                "localtest/public",
                "localtest/otherfs",
                "localtest/noexpire",
            ],
        )
        self.assertEqual(output.col(1), ["True", "True", "False", "False"])
        self.assertEqual(output.col(2), ["Never", "30", "30", "Never"])

        # --yaml
        stdout, _, _ = call_command("workspace", "list", "--yaml")
        data = yaml.safe_load(stdout)
        expected_data = [
            {
                'expiration': 'Never',
                'name': 'debusine/System',
                'public': True,
            },
            {
                'expiration': 30,
                'name': 'localtest/public',
                'public': True,
            },
            {
                'expiration': 30,
                'name': 'localtest/otherfs',
                'public': False,
            },
            {
                'expiration': 'Never',
                'name': 'localtest/noexpire',
                'public': False,
            },
        ]
        self.assertEqual(data, expected_data)

        # scope
        stdout, _, _ = call_command("workspace", "list", "localtest", "--yaml")
        data = yaml.safe_load(stdout)
        self.assertEqual(data, expected_data[1:])

    def test_roles(self) -> None:
        """Test roles commands."""
        self.playground.create_group("Admins", scope=self.scope)
        self.playground.create_group("Employees", scope=self.scope)
        self.playground.create_group("Contractors", scope=self.scope)

        # fmt: off
        call_command("workspace", "create", "localtest/myws", "Admins")

        with self.assertRaisesRegex(
            CommandError,
            "Error assigning role: 'xxx' is not a valid WorkspaceRoles",
        ):
            call_command(
                "workspace", "grant_role", "localtest/myws",
                "xxx", "Employees",
            )

        call_command(
            "workspace", "grant_role", "localtest/myws",
            "contributor", "Employees", "Contractors",
        )

        with self.assertPrintsTable() as output:
            stdout, _, exit_code = call_command(
                "workspace", "list_roles", "localtest/myws"
            )
        self.assertEqual(output.col(0), ["Admins", "Contractors", "Employees"])
        self.assertEqual(output.col(1), ["owner", "contributor", "contributor"])
        self.assertEqual(exit_code, 0)

        stdout, _, exit_code = call_command(
            "workspace", "list_roles", "localtest/myws", "--yaml"
        )
        data = yaml.safe_load(stdout)
        expected_data = [
            {'group': 'Admins', 'role': 'owner'},
            {'group': 'Contractors', 'role': 'contributor'},
            {'group': 'Employees', 'role': 'contributor'},
        ]
        self.assertEqual(data, expected_data)
        self.assertEqual(exit_code, 0)

        # idempotent
        call_command(
            "workspace", "grant_role", "localtest/myws",
            "contributor", "Employees", "Contractors",
        )
        stdout, _, exit_code = call_command(
            "workspace", "list_roles", "localtest/myws", "--yaml"
        )
        data = yaml.safe_load(stdout)
        self.assertEqual(data, expected_data)

        # revoke_role
        stdout, _, exit_code = call_command(
            "workspace", "revoke_role", "localtest/myws",
            "contributor", "Employees"
        )
        stdout, _, exit_code = call_command(
            "workspace", "list_roles", "localtest/myws", "--yaml"
        )
        data = yaml.safe_load(stdout)
        self.assertEqual(data, expected_data[0:-1])
        # fmt: on

    def create_workspace_for_delete(self) -> Workspace:
        """Create a test workspace."""
        scope = self.playground.get_or_create_scope("scope")
        workspace = self.playground.create_workspace(name="Test", scope=scope)
        self.playground.create_group_role(
            workspace,
            Workspace.Roles.CONTRIBUTOR,
            [self.playground.get_default_user()],
        )
        for architecture in ("amd64", "s390x"):
            self.playground.create_debian_environment(
                codename="bookworm",
                architecture=architecture,
                workspace=workspace,
            )
        work_request = self.playground.create_worker_task(
            workspace=workspace, task_name="noop"
        )
        collection = self.playground.create_collection(
            "test", CollectionCategory.WORKFLOW_INTERNAL, workspace=workspace
        )
        artifact_hello, _ = self.playground.create_artifact(
            category=ArtifactCategory.SOURCE_PACKAGE,
            data={
                "name": "hello",
                "version": "1.0-1",
                "type": "dpkg",
                "dsc_fields": {},
            },
            paths=[
                "hello_1.0-1.dsc",
                "hello_1.0-1.debian.tar.xz",
                "hello_1.0.orig.tar.xz",
            ],
            workspace=workspace,
            create_files=True,
            skip_add_files_in_store=True,
        )
        artifact_hello.created_by_work_request = work_request
        artifact_hello.save()

        artifact_hello_traditional, _ = self.playground.create_artifact(
            category=ArtifactCategory.SOURCE_PACKAGE,
            data={
                "name": "hello-traditional",
                "version": "1.0-1",
                "type": "dpkg",
                "dsc_fields": {},
            },
            paths=[
                "hello-traditional.0-1.dsc",
                "hello-traditional.0-1.debian.tar.xz",
                "hello-traditional.0.orig.tar.xz",
            ],
            workspace=workspace,
            create_files=True,
            skip_add_files_in_store=True,
        )
        self.playground.create_artifact_relation(
            artifact_hello_traditional,
            artifact_hello,
            ArtifactRelation.Relations.RELATES_TO,
        )
        collection.manager.add_artifact(
            artifact_hello,
            user=self.playground.get_default_user(),
            name="hello",
        )
        collection.manager.add_artifact(
            artifact_hello_traditional,
            user=self.playground.get_default_user(),
            name="hello-traditional",
        )
        WorkflowTemplate.objects.create(
            name="test", workspace=workspace, task_name="noop"
        )

        sbuild_template = self.playground.create_workflow_template(
            name="Build package",
            task_name="sbuild",
            static_parameters={},
            workspace=workspace,
        )

        udev = self.playground.create_source_artifact(
            name="udev",
            version="252.26-1~deb12u2",
            create_files=True,
            workspace=workspace,
        )

        workflow = self.playground.create_workflow(
            sbuild_template,
            task_data={
                "input": {
                    "source_artifact": udev.pk,
                },
                "target_distribution": "debian:bookworm",
                "architectures": ["all", "amd64", "s390x"],
            },
        )
        workflow.mark_running()
        workflow.save()

        assert workflow.internal_collection is not None
        self.assertEqual(workflow.internal_collection.workflow, workflow)

        # TODO: add more kinds of elements to the workspace, to make
        # sure deletion catches them
        return workspace

    def test_delete_workspace(self) -> None:
        """Delete an existing workspace."""
        workspace = self.create_workspace_for_delete()

        stdout, _, _ = call_command(
            "workspace", "delete", "scope/Test", "--yes"
        )
        self.assertEqual("", stdout)

        with self.assertRaises(Workspace.DoesNotExist):
            Workspace.objects.get(id=workspace.id)

        with self.assertRaises(Workspace.DoesNotExist):
            Workspace.objects.get(name="Test")

    def test_delete_workspace_missing(self) -> None:
        """Try to delete a workspace that does not exist."""
        self.playground.get_or_create_scope("scope")
        with self.assertRaisesRegex(
            CommandError, r"Workspace 'Test' not found in scope 'scope'$"
        ) as exc:
            call_command("workspace", "delete", "scope/Test", "--yes")
        self.assertEqual(exc.exception.returncode, 3)

        stdout, stderr, exit_code = call_command(
            "workspace",
            "delete",
            "scope/Test",
            "--force",
        )
        self.assertEqual(stdout, "")
        self.assertEqual(stderr, "")
        self.assertEqual(exit_code, 0)

        stdout, stderr, exit_code = call_command(
            "workspace",
            "delete",
            "scope/Test",
            "--force",
            "--yes",
        )
        self.assertEqual(stdout, "")
        self.assertEqual(stderr, "")
        self.assertEqual(exit_code, 0)

    def test_delete_workspace_playground_ui_scenario(self) -> None:
        """Delete a workspace with the playground UI scenario."""
        playground = Playground(default_workspace_name="Playground")
        scenario = scenarios.UIPlayground()
        playground.build_scenario(scenario)

        stdout, _, _ = call_command(
            "workspace", "delete", str(scenario.workspace), "--yes"
        )
        self.assertEqual("", stdout)

        with self.assertRaises(Workspace.DoesNotExist):
            Workspace.objects.get(id=scenario.workspace.id)

        with self.assertRaises(Workspace.DoesNotExist):
            Workspace.objects.get(
                scope=scenario.scope, name=scenario.workspace.name
            )

    def test_delete_workspace_confirmation(self) -> None:
        """delete_workspace doesn't delete (user does not confirm)."""
        workspace = self.create_workspace_for_delete()

        call_command(
            "workspace",
            "delete",
            "scope/Test",
            stdin=io.StringIO("N\n"),
        )
        self.assertQuerySetEqual(
            Workspace.objects.filter(name="Test"), [workspace]
        )

        call_command(
            "workspace",
            "delete",
            "scope/Test",
            stdin=io.StringIO("\n"),
        )
        self.assertQuerySetEqual(
            Workspace.objects.filter(name="Test"), [workspace]
        )

    def test_delete_default_workspace(self) -> None:
        """Default workspace does not get deleted."""
        workspace = default_workspace()
        with self.assertRaisesRegex(
            CommandError,
            rf"^Workspace {workspace.scope.name}/{workspace.name}"
            " cannot be deleted$",
        ) as exc:
            call_command(
                "workspace",
                "delete",
                str(workspace),
                "--yes",
            )
        self.assertEqual(exc.exception.returncode, 3)



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""debusine signon URLs Configuration."""

from django.urls import path

from debusine.server.signon.views import (
    BindIdentityView,
    OIDCAuthenticationCallbackView,
)

app_name = "signon"

urlpatterns = [
    path(
        "oidc_callback/<name>/",
        OIDCAuthenticationCallbackView.as_view(),
        name="oidc_callback",
    ),
    path(
        "bind_identity/<name>/",
        BindIdentityView.as_view(),
        name="bind_identity",
    ),
]



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""debusine celery configuration."""

import logging
import os
import secrets
from collections.abc import Callable, Sequence
from functools import partial
from pathlib import Path
from types import SimpleNamespace
from typing import Any, TYPE_CHECKING
from uuid import uuid4

from celery import Celery, Task, states
from celery.result import AsyncResult
from celery.signals import before_task_publish, worker_init, worker_shutdown
from celery.worker import WorkController
from django.db import connection, transaction
from django.db.models import IntegerField, JSONField, QuerySet
from django.db.models.functions import Cast

from debusine.client.debusine import Debusine

if TYPE_CHECKING:
    from django_celery_results.models import TaskResult

    from debusine.db.models import Token

os.environ.setdefault("DJANGO_SETTINGS_MODULE", "debusine.project.settings")


# mypy complains that Task is untyped, which is true, but we can't fix that
# here.
class TaskWithBetterArgumentRepresentation(Task):  # type: ignore[misc]
    """
    Custom task class to control result argument serialization.

    See https://github.com/celery/django-celery-results/issues/113.
    """

    def apply_async(
        self,
        args: tuple[Any, ...] | None = None,
        kwargs: dict[str, Any] | None = None,
        **options: Any,
    ) -> AsyncResult:
        """
        Override Celery's default argument encoding in the result backend.

        By default, Celery encodes arguments using ``repr()`` when storing
        them in the result backend, which makes it difficult to query them.
        Passing them straight through allows ``django_celery_results`` to
        apply JSON encoding instead.
        """
        options["argsrepr"] = args or ()
        options["kwargsrepr"] = kwargs or {}
        return super().apply_async(args=args, kwargs=kwargs, **options)


def make_app() -> Celery:
    """Make a Celery application."""
    app = Celery("debusine", task_cls=TaskWithBetterArgumentRepresentation)
    app.config_from_object("django.conf:settings", namespace="CELERY")
    app.autodiscover_tasks()
    return app


app = make_app()


# Based loosely on:
#   https://stackoverflow.com/q/9824172
#   https://github.com/celery/django-celery-results/issues/286
# mypy complains that before_task_publish.connect is untyped, which is true,
# but we can't fix that here.
@before_task_publish.connect  # type: ignore[untyped-decorator]
def store_pending_state(
    *, sender: str, headers: dict[str, Any], **kwargs: Any
) -> None:
    """Tell the result backend more details about newly-published tasks."""
    task = app.tasks.get(sender)
    if task is not None:  # pragma: no cover
        # Since we need to tell the result backend about tasks before they
        # are created (in apply_async below), we don't support individual
        # tasks using a different backend.
        assert task.backend == app.backend
    app.backend.store_result(
        headers["id"], None, states.PENDING, request=SimpleNamespace(**headers)
    )


def _apply_async_on_commit_callback(
    task: Task,
    *,
    args: tuple[Any] | None = None,
    kwargs: dict[Any, Any] | None = None,
    task_id: str | None = None,
    **options: Any,
) -> None:
    from django_celery_results.models import TaskResult

    task.apply_async(args=args, kwargs=kwargs, task_id=task_id, **options)
    if app.conf.task_always_eager:
        # In always-eager mode, the TaskResult row won't be automatically
        # updated after the task has completed, but we have no further use
        # for it anyway.  Delete it so that code that inspects the task
        # queue doesn't think the task is still pending.
        TaskResult.objects.filter(task_id=task_id).delete()


def get_current_savepoint_run_on_commit_callbacks() -> list[Callable[[], None]]:
    """Get scheduled ``run_on_commit`` callbacks for the current savepoint."""
    # This uses undocumented Django internals, but if it goes wrong without
    # an obvious unit test failure, then the worst case should be some extra
    # calls to a function that we normally expect to be batched, which is
    # relatively harmless.
    return [
        callback
        for savepoint_ids, callback, *_ in connection.run_on_commit
        if savepoint_ids == set(connection.savepoint_ids)
    ]


def get_apply_async_tasks(
    callbacks: Sequence[Callable[[], None]],
) -> list[Task]:
    """Get Celery tasks scheduled to be run on commit."""
    return [
        callback.args[0]
        for callback in callbacks
        if isinstance(callback, partial)
        and callback.func == _apply_async_on_commit_callback
    ]


def apply_async_on_commit(
    task: Task,
    *,
    args: tuple[Any] | None = None,
    kwargs: dict[Any, Any] | None = None,
    task_id: str | None = None,
    once_per_savepoint: bool = False,
    **options: Any,
) -> str | None:
    """
    Apply a task asynchronously when the transaction is committed.

    Tell the result backend about it even before the transaction is
    committed, so that code in the same transaction can tell that the task
    is pending.

    :return: if a task was created, the Celery task ID; otherwise None.
    """
    # Some tasks should be batched up and only run once per savepoint.
    if once_per_savepoint:
        assert args is None
        assert kwargs is None
        if task in get_apply_async_tasks(
            get_current_savepoint_run_on_commit_callbacks()
        ):
            return None

    if task_id is None:
        task_id = str(uuid4())

    app.backend.store_result(
        task_id,
        None,
        states.PENDING,
        # This is just enough for code that needs to inspect the task queue.
        request=SimpleNamespace(
            task=task.name, argsrepr=args or (), kwargsrepr=kwargs or {}
        ),
    )
    transaction.on_commit(
        partial(
            _apply_async_on_commit_callback,
            task,
            args=args,
            kwargs=kwargs,
            task_id=task_id,
            **options,
        )
    )

    return task_id


def celery_tasks_in_progress(task_name: str) -> QuerySet["TaskResult"]:
    """Return a QuerySet of Celery tasks in progress with a given task name."""
    from django_celery_results.models import TaskResult

    task_results = TaskResult.objects.filter(
        status__in={
            states.PENDING,
            states.RECEIVED,
            states.STARTED,
            states.RETRY,
        },
        task_name=task_name,
    ).annotate(
        task_args_array=Cast("task_args", JSONField()),
        first_task_arg=Cast("task_args_array__0", IntegerField()),
    )
    assert isinstance(task_results, QuerySet)
    return task_results


class CeleryToken:
    """Manage a Celery worker's token."""

    token: "Token"
    client: Debusine

    def __init__(self, name: str) -> None:
        """Read an existing token key, or generate one if there is none."""
        from django.conf import settings

        from debusine.db.models import Token
        from debusine.utils import atomic_writer

        path = Path(settings.DEBUSINE_TOKENS_DIRECTORY, name)
        try:
            key = path.read_text()
        except FileNotFoundError:
            key = secrets.token_hex(32)
            with atomic_writer(path, mode="w", chmod=0o600) as fd:
                fd.write(key)

        self.token, _ = Token.objects.get_or_create(
            token_type=Token.Types.WORKER, hash=Token._generate_hash(key)
        )
        self.client = Debusine(
            f"https://{settings.DEBUSINE_FQDN}/api",
            key,
            logger=logging.getLogger(),
        )


_celery_token: CeleryToken | None = None


def get_celery_token() -> CeleryToken | None:
    """Get information about the current Celery worker token."""
    return _celery_token


def get_celery_worker_provided_tags() -> set[str]:
    """Return a set with tags provided by the Celery worker."""
    from debusine.server.tasks import get_provided_celery_worker_tags
    from debusine.tasks.models import WorkerType
    from debusine.worker.system_information import system_provided_tags

    tags: set[str] = set()
    tags |= system_provided_tags(WorkerType.CELERY)
    tags |= get_provided_celery_worker_tags()
    return tags


def get_celery_worker_metadata() -> dict[str, Any]:
    """Compute the dynamic metadata sent by the Celery worker."""
    from debusine.server.tasks import analyze_celery_worker_tasks
    from debusine.tasks.models import WorkerType
    from debusine.worker.system_information import system_metadata

    return {
        **system_metadata(WorkerType.CELERY),
        **analyze_celery_worker_tasks(),
        "provided_tags": sorted(get_celery_worker_provided_tags()),
    }


def get_scheduler_worker_metadata() -> dict[str, Any]:
    """Compute the dynamic metadata sent by the scheduler worker."""
    from debusine.tasks.models import WorkerType
    from debusine.worker.system_information import system_metadata

    return {**system_metadata(WorkerType.SCHEDULER)}


# mypy complains that worker_init.connect is untyped, which is true, but we
# can't fix that here.
@worker_init.connect  # type: ignore[untyped-decorator]
@transaction.atomic
def connect_worker(*, sender: WorkController, **kwargs: Any) -> None:
    """Set up and connect a :py:class:`Worker` when Celery starts."""
    from debusine.db.models import Worker

    worker: Worker | None = None

    if sender.hostname.startswith("celery@"):
        worker = Worker.objects.get_or_create_celery()
        worker.concurrency = sender.concurrency
        worker.save()
        metadata = get_celery_worker_metadata()
    elif sender.hostname.startswith("scheduler@"):
        worker = Worker.objects.get_or_create_scheduler()
        metadata = get_scheduler_worker_metadata()

    if worker is not None:
        global _celery_token
        _celery_token = CeleryToken(sender.hostname.split("@")[0])
        if worker.token is None:
            worker.token = _celery_token.token
        worker.save()
        worker.token.enable()

        worker.set_dynamic_metadata(metadata)
        worker.mark_connected()


# mypy complains that worker_shutdown.connect is untyped, which is true, but
# we can't fix that here.
@worker_shutdown.connect  # type: ignore[untyped-decorator]
def disconnect_worker(*, sender: WorkController, **kwargs: Any) -> None:
    """Disconnect the appropriate :py:class:`Worker` when Celery exits."""
    from debusine.db.models import Worker

    if sender.hostname.startswith("celery@"):
        Worker.objects.get_or_create_celery().mark_disconnected()
    elif sender.hostname.startswith("scheduler@"):
        Worker.objects.get_or_create_scheduler().mark_disconnected()

    global _celery_token
    _celery_token = None



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""debusine-admin command to print server configuration information."""

import shutil
import socket
import textwrap
from typing import Any, NoReturn, override

from django.conf import settings

from debusine.django.management.debusine_base_command import DebusineBaseCommand


class Command(DebusineBaseCommand):
    """Command to show server configuration."""

    help = "Show server configuration to help deploy"

    @override
    def handle(self, *args: Any, **options: Any) -> NoReturn:
        """Show server configuration."""
        cols, _ = shutil.get_terminal_size()
        wrapper = textwrap.TextWrapper(width=cols)

        def wprint(text: str) -> None:
            self.stdout.write("\n".join(wrapper.wrap(text)))

        self.stdout.write()
        self.stdout.write(f" * DEBUSINE_FQDN = {settings.DEBUSINE_FQDN!r}")
        self.stdout.write()
        wprint(
            "DEBUSINE_FQDN is the hostname used to generate URLs to debusine."
            " It can be different from the machine's hostname in case debusine"
            " is running, for example, on a server behind a proxy or load"
            " balancer."
        )
        self.stdout.write()
        socket_fqdn = socket.getfqdn()
        if socket_fqdn != settings.DEBUSINE_FQDN:
            wprint(
                f"The current value is changed from the default {socket_fqdn!r}"
            )
        else:
            wprint(
                "The current value matches the autodetected default"
                " from socket.getfqdn()."
            )

        self.stdout.write()
        self.stdout.write(f" * ALLOWED_HOSTS = {settings.ALLOWED_HOSTS!r}")
        self.stdout.write()
        wprint(
            "See https://docs.djangoproject.com/en/5.0/ref/settings/#allowed-hosts"  # noqa: E501
        )
        self.stdout.write()
        allowed_hosts_default = [
            settings.DEBUSINE_FQDN,
            "localhost",
            "127.0.0.1",
        ]
        if settings.ALLOWED_HOSTS != allowed_hosts_default:
            wprint(
                "The current value is changed from default"
                f" {allowed_hosts_default!r}"
            )
        else:
            wprint(
                "The current value matches the autodetected default from"
                " DEBUSINE_FQDN."
            )
        raise SystemExit(0)



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Tests for the management command create_token."""

from django.contrib.auth import get_user_model
from django.core.management import CommandError
from django.test import TestCase

from debusine.db.models import Token
from debusine.django.management.tests import call_command


class CreateTokenCommandTests(TestCase):
    """Tests for the create_token command."""

    def test_create_token(self) -> None:
        """create_token creates a new token and prints the key on stdout."""
        username = "james"
        get_user_model().objects.create_user(username=username, password="1234")
        stdout, stderr, exit_code = call_command('create_token', username)

        token = Token.objects.latest("id")

        # The token key is not stored in the db, so we need to hash the token
        # we got.
        token_hash = Token._generate_hash(stdout.strip())

        self.assertEqual(token_hash, token.hash)
        self.assertEqual(stderr, "")
        self.assertEqual(exit_code, 0)
        self.assertTrue(token.enabled)

    def test_username_does_not_exist(self) -> None:
        """create_token cannot create a token: username does not exist."""
        username = "james"
        expected_error = f'Cannot create token: "{username}" does not exist'
        with self.assertRaisesRegex(CommandError, expected_error) as exc:
            call_command("create_token", username)

        self.assertEqual(exc.exception.returncode, 3)
        self.assertEqual(Token.objects.count(), 0)



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Unit tests for the makesourcepackageupload task support on the worker."""

from pathlib import Path
from typing import override
from unittest import mock
from unittest.mock import call

import debusine.tasks.tags as ttags
import debusine.worker.tags as wtags
from debusine.artifacts import Upload
from debusine.artifacts.models import (
    ArtifactCategory,
    DebianSourcePackage,
    WorkRequestResults,
)
from debusine.client.models import LookupSingleResponseArtifact
from debusine.tasks import MakeSourcePackageUpload, TaskConfigError
from debusine.tasks.models import MakeSourcePackageUploadDynamicData
from debusine.tasks.tests.helpers import ExternalTaskHelperMixin
from debusine.test import TestCase
from debusine.test.test_utils import (
    create_artifact_response,
    create_remote_artifact,
    create_system_tarball_data,
)


class MakeSourcePackageUploadTests(
    ExternalTaskHelperMixin[MakeSourcePackageUpload], TestCase
):
    """Test the MakeSourcePackageUpload class."""

    SAMPLE_TASK_DATA = {
        "environment": "debian/match:codename=bookworm",
        "input": {"source_artifact": 421},
    }

    @override
    def setUp(self) -> None:
        super().setUp()
        self.resolver.set_field_artifact(
            "environment",
            id=1,
            category=ArtifactCategory.SYSTEM_TARBALL,
            data=create_system_tarball_data(),
        )
        self.resolver.set_field_artifact(
            "input_source_artifact",
            id=421,
            category=ArtifactCategory.SOURCE_PACKAGE,
            data=DebianSourcePackage(
                name="hello", version="1.0", type="dpkg", dsc_fields={}
            ),
        )

    def test_get_provided_worker_tags(self) -> None:
        self.assertCountEqual(
            self.task.get_provided_worker_tags(),
            [
                wtags.TASK_PREFIX + "worker:makesourcepackageupload:version:1",
            ],
        )

    def test_compute_scheduler_tags(self) -> None:
        self.assertComputesSchedulerTags(
            provide_system={ttags.SOURCE_PACKAGE_PREFIX + "hello"},
            require_system={
                wtags.BUILD_ARCH_PREFIX + "amd64",
                wtags.EXECUTOR_PREFIX + "unshare",
            },
        )

    def test_compute_dynamic_data(self) -> None:
        """Dynamic data receives relevant artifact IDs."""
        self.assertEqual(
            self.compute_dynamic_data(self.task),
            MakeSourcePackageUploadDynamicData(
                environment_id=1,
                input_source_artifact_id=421,
                subject="hello",
                parameter_summary="hello_1.0",
            ),
        )

    def test_get_subject(self) -> None:
        self.assertEqual(self.get_subject(), "hello")

    def test_get_configuration_context(self) -> None:
        self.assertIsNone(self.get_configuration_context())

    def test_acceptable_input_categories(self) -> None:
        self.assertEqual(
            MakeSourcePackageUpload.inputs["input_source_artifact"].categories,  # type: ignore[attr-defined]
            (ArtifactCategory.SOURCE_PACKAGE,),
        )

    def test_get_input_artifacts_ids(self) -> None:
        """Test get_input_artifacts_ids."""
        self.assertEqual(self.task.get_input_artifacts_ids(), [])

        self.task.dynamic_data = MakeSourcePackageUploadDynamicData(
            environment_id=1,
            input_source_artifact_id=2,
        )
        self.assertEqual(self.task.get_input_artifacts_ids(), [1, 2])

    def test_configure_fails_with_missing_required_data(
        self,
    ) -> None:
        with self.assertRaises(TaskConfigError):
            self.configure_task(override={"input": {}})

    def test_fetch_input(self) -> None:
        """Test fetch_input: call fetch_artifact(artifact_id, directory)."""
        directory = self.create_temporary_directory()
        source_artifact = self.fake_debian_source_package_artifact()
        self.configure_task()
        self.task.work_request_id = 5
        self.task.dynamic_data = MakeSourcePackageUploadDynamicData(
            environment_id=1, input_source_artifact_id=source_artifact.id
        )
        debusine_mock = self.mock_debusine()
        debusine_mock.artifact_get.return_value = source_artifact

        with mock.patch.object(
            self.task, "fetch_artifact", autospec=True, return_value=True
        ) as fetch_artifact_mocked:
            result = self.task.fetch_input(directory)

        self.assertTrue(result)
        fetch_artifact_mocked.assert_called_once_with(
            source_artifact.id, directory
        )

    def test_fetch_input_wrong_category(self) -> None:
        """Test fetch_input when input isn't a source package."""
        directory = self.create_temporary_directory()
        source_artifact = create_artifact_response(id=12)
        self.configure_task()
        self.task.work_request_id = 5
        self.task.dynamic_data = MakeSourcePackageUploadDynamicData(
            environment_id=1, input_source_artifact_id=source_artifact.id
        )
        debusine_mock = self.mock_debusine()
        debusine_mock.artifact_get.return_value = source_artifact

        with mock.patch.object(
            self.task, "fetch_artifact", autospec=True
        ) as fetch_artifact_mocked:
            result = self.task.fetch_input(directory)

        self.assertFalse(result)

        log_file_contents = (
            self.task.log_directory / "fetch_input.log"
        ).read_text()
        self.assertEqual(
            log_file_contents,
            (
                "input.source_artifact points to a Testing, not the "
                "expected debian:source-package.\n"
            ),
        )

        fetch_artifact_mocked.assert_not_called()

    def test_configure_for_execution(self) -> None:
        """configure_for_execution installs dependencies."""
        download_directory = self.create_temporary_directory()
        (download_directory / "file.dsc").touch()
        mocked_prepare_executor_instance = (
            self.patch_prepare_executor_instance()
        )

        self.assertTrue(self.task.configure_for_execution(download_directory))

        mocked_prepare_executor_instance.assert_called_once_with()
        assert isinstance(self.task.executor_instance, mock.MagicMock)
        self.task.executor_instance.assert_has_calls(
            [
                mock.call.run(
                    ["apt-get", "update"], run_as_root=True, check=True
                ),
                mock.call.run(
                    ["apt-get", "--yes", "install", "dpkg-dev"],
                    run_as_root=True,
                    check=True,
                ),
            ]
        )

    def test_execute(self) -> None:
        """Test full (mocked) execution."""
        self.configure_task()
        self.task.work_request_id = 2
        self.task.workspace_name = "testing"
        self.task.dynamic_data = MakeSourcePackageUploadDynamicData(
            environment_id=1, input_source_artifact_id=1
        )
        download_directory = self.create_temporary_directory()

        debusine_mock = self.mock_debusine()
        debusine_mock.lookup_single.return_value = LookupSingleResponseArtifact(
            artifact=1
        )
        debusine_mock.download_artifact.return_value = True
        debusine_mock.upload_artifact.return_value = create_remote_artifact(
            id=2, workspace=self.task.workspace_name
        )

        f_in_contents = "Format: 3.0 (quilt)"
        (f_in := download_directory / "file.dsc").write_text(f_in_contents)
        (f_out := download_directory / "file_source.changes")
        self.write_changes_file(f_out, [f_in])

        self.patch_prepare_executor_instance()

        self.assertTrue(self.task.configure_for_execution(download_directory))

        self.assertEqual(self.task._dsc_path, f_in)
        self.assertEqual(
            self.task._cmdline(),
            [
                "bash",
                "-x",
                "-e",
                str(download_directory / "makesourcepackageupload.sh"),
                str(f_in),
                str(f_out),
                '',
                '',
            ],
        )

        self.task.upload_artifacts(
            download_directory, execution_result=WorkRequestResults.SUCCESS
        )

        debusine_mock.upload_artifact.assert_called_once_with(
            Upload.create(changes_file=f_out),
            workspace=self.task.workspace_name,
            work_request=self.task.work_request_id,
        )
        debusine_mock.relation_create.assert_has_calls(
            [mock.call(2, 1, "extends"), mock.call(2, 1, "relates-to")]
        )

    def test_optional_arguments(self) -> None:
        """Test optional arguments."""
        self.configure_task(override={"since_version": "2.10-1~"})
        self.task._shell_script = Path("script")
        self.task._dsc_path = Path("in")
        self.task._changes_path = Path("out")
        expected = ["bash", "-x", "-e", "script", "in", "out"]
        self.assertEqual(
            self.task._cmdline(),
            [*expected, "2.10-1~", ""],
        )

        self.configure_task(
            override={"target_distribution": "bullseye-security"}
        )
        self.task._shell_script = Path("script")
        self.task._dsc_path = Path("in")
        self.task._changes_path = Path("out")
        self.assertEqual(
            self.task._cmdline(),
            [*expected, "", "bullseye-security"],
        )

        self.configure_task(
            override={
                "since_version": "1 0",
                "target_distribution": "buster elts",
            }
        )
        self.task._shell_script = Path("script")
        self.task._dsc_path = Path("in")
        self.task._changes_path = Path("out")
        self.assertEqual(
            self.task._cmdline(),
            [*expected, "1 0", "buster elts"],
        )

    def test_upload_artifacts(self) -> None:
        """upload_artifact() and relation_create() is called."""
        self.task.dynamic_data = MakeSourcePackageUploadDynamicData(
            environment_id=1, input_source_artifact_id=1
        )
        download_directory = self.create_temporary_directory()

        # Create file that will be attached when uploading the artifacts
        f_in_contents = "Format: 3.0 (quilt)"
        (f_in := download_directory / "file.dsc").write_text(f_in_contents)
        (f_out := download_directory / "file_source.changes")
        self.write_changes_file(f_out, [f_in])
        self.task._changes_path = f_out

        # Debusine.upload_artifact is mocked to verify the call only
        debusine_mock = self.mock_debusine()

        workspace_name = "testing"

        uploaded_artifacts = [
            create_remote_artifact(id=10, workspace=workspace_name),
        ]

        debusine_mock.upload_artifact.side_effect = uploaded_artifacts

        # self.task.workspace_name is set by the Worker
        # and is the workspace that downloads the artifact
        # containing the files needed for MakeSourcePackageUpload
        self.task.workspace_name = workspace_name

        # The worker set self.task.work_request_id of the task
        work_request_id = 147
        self.task.work_request_id = work_request_id

        with mock.patch.object(self.task, "executor_instance", autospec=True):
            self.task.upload_artifacts(
                download_directory, execution_result=WorkRequestResults.SUCCESS
            )

        # Debusine Mock upload_artifact expected calls
        upload_artifact_calls = []
        makesourcepackageupload_artifact = Upload.create(changes_file=f_out)

        upload_artifact_calls.append(
            call(
                makesourcepackageupload_artifact,
                workspace=workspace_name,
                work_request=work_request_id,
            )
        )

        # Debusine mock relation_create expected calls
        relation_create_calls = []
        relation_create_calls.append(
            call(uploaded_artifacts[0].id, 1, "relates-to")
        )

        # Assert that the artifacts were uploaded and relations created
        debusine_mock.upload_artifact.assert_has_calls(upload_artifact_calls)
        debusine_mock.relation_create.assert_has_calls(relation_create_calls)



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Tests for archives templatetags."""

from typing import override
from unittest import mock

from django.template import TemplateSyntaxError, engines
from django.test import RequestFactory, override_settings
from django.urls import get_urlconf, reverse, set_urlconf

from debusine.db.context import context
from debusine.db.playground import scenarios
from debusine.test.django import TestCase


class DebusineURLTests(TestCase):
    """Tests for debusine_url tag."""

    # The debusine_url template tag currently has dependencies on Django
    # internals, so code coverage is especially important here, to try and
    # check all possible tag behaviours and be more likely to detect breakage
    # with future Django versions

    scenario = scenarios.DefaultContext(set_current=True)

    fqdn = "debusine.example.org"

    @override
    def setUp(self) -> None:
        super().setUp()
        self.enterContext(
            override_settings(
                DEBUSINE_URL_SCHEME="https", DEBUSINE_FQDN=self.fqdn
            )
        )

    def render(self, template_code: str) -> str:
        """Render a template from a string."""
        orig = get_urlconf()
        try:
            set_urlconf("debusine.web.archives.urls")

            request = RequestFactory().get("/")
            template = engines["django"].from_string(
                "{% load archives %}" + template_code
            )
            return template.render({"context": context}, request=request)
        finally:
            set_urlconf(orig)

    def test_homepage(self) -> None:
        self.assertEqual(
            self.render(
                "{% debusine_url 'homepage:homepage' %}",
            ),
            f"https://{self.fqdn}" + reverse("homepage:homepage"),
        )

    def test_workspace(self) -> None:
        self.assertEqual(
            self.render(
                f"{{% debusine_url 'workspaces:detail' "
                f"wname={self.scenario.workspace.name!r} %}}",
            ),
            f"https://{self.fqdn}"
            + reverse(
                "workspaces:detail",
                kwargs={"wname": self.scenario.workspace.name},
            ),
        )

    def test_workspace_positional(self) -> None:
        self.assertEqual(
            self.render(
                f"{{% debusine_url 'workspaces:detail' "
                f"{self.scenario.workspace.name!r} %}}",
            ),
            f"https://{self.fqdn}"
            + reverse(
                "workspaces:detail",
                kwargs={"wname": self.scenario.workspace.name},
            ),
        )

    def test_as(self) -> None:
        self.assertEqual(
            self.render(
                "{% debusine_url 'homepage:homepage' as test %}{{test}}",
            ),
            f"https://{self.fqdn}" + reverse("homepage:homepage"),
        )

    def test_argument_required(self) -> None:
        with self.assertRaisesRegex(
            TemplateSyntaxError,
            "takes at least one argument, a URL pattern name.",
        ):
            self.render(
                "{% debusine_url %}",
            )

    def test_assignment_syntax_check(self) -> None:
        with (
            # The way django.template.base.kwarg_re is currently set, there
            # does not seem to be a string that could be generated by the
            # template tokenizer and would not match the regexp.
            #
            # This may change if in the future kwarg_re gets changed, so here's
            # a mock to keep that code path exercised
            mock.patch(
                "django.template.base.Token.split_contents",
                return_value=["url", "'homepage:homepage'", ""],
            ),
            self.assertRaisesRegex(
                TemplateSyntaxError, "Malformed arguments to url tag"
            ),
        ):
            self.render(
                "{% debusine_url 'homepage:homepage' %}",
            )



#  Copyright © The Debusine Developers
#  See the AUTHORS file at the top-level directory of this distribution
#
#  This file is part of Debusine. It is subject to the license terms
#  in the LICENSE file found in the top-level directory of this
#  distribution. No part of Debusine, including this file, may be copied,
#  modified, propagated, or distributed except according to the terms
#  contained in the LICENSE file.

"""Unit tests for manage.py."""

import contextlib
import io
from unittest import mock

from django.core.exceptions import ImproperlyConfigured

from debusine.__main__ import main
from debusine.test.django import TestCase


class DefaultSettingsTests(TestCase):
    """Tests for manage.py. Named debusine-admin when packaged."""

    def patch_execute_from_command_line(self, exception: BaseException) -> None:
        """Patch execute_from_command_line: raise exception."""
        patcher = mock.patch(
            'django.core.management.execute_from_command_line', autospec=True
        )
        mocked_pathlib_owner = patcher.start()
        mocked_pathlib_owner.side_effect = exception
        self.addCleanup(patcher.stop)

    def assert_main_drops_privileges(
        self, program_name: str, target_username: str
    ) -> None:
        """Call main(), asserting that it drops privileges."""
        with (
            mock.patch("sys.argv", [program_name, "--version"]),
            mock.patch("os.geteuid", return_value=0),
            mock.patch("os.execvp", side_effect=OSError) as mock_execvp,
            mock.patch("debusine.django.django_utils.test_data_override"),
            self.assertRaises(OSError),
        ):
            main()

        mock_execvp.assert_called_once_with(
            "runuser",
            [
                "runuser",
                f"--user={target_username}",
                "--",
                program_name,
                "--version",
            ],
        )

    def assert_main_does_not_drop_privileges(self, program_name: str) -> None:
        """Call main(), asserting that it doesn't drop privileges."""
        stdout = io.StringIO()

        with (
            mock.patch("sys.argv", [program_name, "--version"]),
            mock.patch("os.geteuid", return_value=1),
            mock.patch("os.execvp", side_effect=OSError) as mock_execvp,
            mock.patch("debusine.django.django_utils.test_data_override"),
            contextlib.redirect_stdout(stdout),
        ):
            main()

        mock_execvp.assert_not_called()

    def assert_main_system_exit_with_stderr(
        self, stderr_expected: str, program_name: str = "debusine-admin"
    ) -> None:
        """Call main() assert system exit code is 3 and stderr as expected."""
        stderr = io.StringIO()

        with (
            mock.patch("sys.argv", [program_name]),
            mock.patch("os.geteuid", return_value=1),
            contextlib.redirect_stderr(stderr),
            self.assertRaisesSystemExit(3),
        ):
            main()

        self.assertEqual(stderr.getvalue(), stderr_expected)

    def test_debusine_admin_as_root_drops_privileges(self) -> None:
        """main() drops privs if running debusine-admin as root."""
        self.assert_main_drops_privileges("debusine-admin", "debusine-server")

    def test_debusine_admin_as_non_root_does_not_drop_privileges(self) -> None:
        """main() doesn't drop privs if running debusine-admin as non-root."""
        self.assert_main_does_not_drop_privileges("debusine-admin")

    def test_debusine_signing_as_root_drops_privileges(self) -> None:
        """main() drops privs if running debusine-signing as root."""
        self.assert_main_drops_privileges(
            "debusine-signing", "debusine-signing"
        )

    def test_debusine_signing_as_non_root_does_not_drop_privileges(
        self,
    ) -> None:
        """main() doesn't drop privs if running debusine-signing as non-root."""
        self.assert_main_does_not_drop_privileges("debusine-signing")

    def test_improperly_configured_output(self) -> None:
        """main() handle ImproperlyConfigured from execute_from_command_line."""
        self.patch_execute_from_command_line(ImproperlyConfigured("Error msg"))

        self.assert_main_system_exit_with_stderr(
            "Improperly configured error: Error msg\n"
        )

    def test_raise_error(self) -> None:
        """main() raise non-handled exception."""
        self.patch_execute_from_command_line(OSError("Some error"))

        with (
            mock.patch("sys.argv", ["debusine-admin", "--version"]),
            mock.patch("os.geteuid", return_value=1),
            self.assertRaises(OSError),
        ):
            main()

    def assert_raise_error_permission_error(
        self, permission_error: PermissionError, msg: str
    ) -> None:
        """Assert that permission_error is raised msg is in the stderr."""
        exception = ValueError("Some error")
        exception.__cause__ = permission_error

        self.patch_execute_from_command_line(exception)

        self.assert_main_system_exit_with_stderr(msg)

    def test_permission_error_raised_unknown_file(self) -> None:
        """main() handle ValueError caused by PermissionError."""
        permission_error = PermissionError("Internal error")

        self.assert_raise_error_permission_error(
            permission_error,
            "Permission error: Some error. Check that the user running "
            'debusine-admin has access to the file "Unknown".\n',
        )

    def test_permission_error_raised_file(self) -> None:
        """main() handle ValueError caused by PermissionError."""
        permission_error = PermissionError("Internal error")

        permission_error.filename = "/var/log/debusine/server/debug.log"

        self.assert_raise_error_permission_error(
            permission_error,
            "Permission error: Some error. Check that the user running "
            "debusine-admin has access to the file "
            f'"{permission_error.filename}".\n',
        )

    def test_value_error(self) -> None:
        """main() handles ValueError not caused by PermissionError."""
        self.patch_execute_from_command_line(ValueError("Some error"))

        with self.assertRaisesRegex(ValueError, "Some error"):
            main()



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Tests for client task configuration handling."""

import logging
import os
import shutil
import subprocess
import time as tm
from collections.abc import Collection
from pathlib import Path
from typing import Any, ClassVar, override
from unittest import SkipTest, mock
from unittest._log import _LoggingWatcher

import pydantic
import yaml

from debusine.artifacts.models import DebusineTaskConfiguration, TaskTypes
from debusine.client.models import TaskConfigurationCollection
from debusine.client.task_configuration import (
    EntryFile,
    GitTaskConfigurationRepository,
    InvalidRepository,
    Item,
    LocalTaskConfigurationRepository,
    Manifest,
    PullStats,
    RemoteTaskConfigurationRepository,
    as_item,
)
from debusine.test import TestCase


class Samples:
    """
    Simple holder for task configuration samples.

    You can assign to members and they will be preserved.

    Accessing a nonexistent member will create a unique one on the fly.
    """

    def add_config(
        self,
        name: str,
        *,
        task_type: TaskTypes = TaskTypes.WORKER,
        task_name: str = "noop",
        **kwargs: Any,
    ) -> DebusineTaskConfiguration:
        """Add a DebusineTaskConfiguration with known values."""
        entry = DebusineTaskConfiguration(
            task_type=task_type, task_name=task_name, **kwargs
        )
        setattr(self, name, entry)
        return entry

    def __getattr__(self, name: str) -> DebusineTaskConfiguration:
        """Instantiate a DebusineTaskConfiguration entry on the fly."""
        result = DebusineTaskConfiguration(template=name)
        setattr(self, name, result)
        return result


def items(*samples: DebusineTaskConfiguration) -> list[Item]:
    """Make a list of items from samples."""
    return [as_item(s) for s in samples]


def with_path(
    sample: DebusineTaskConfiguration, path: str, position: int
) -> DebusineTaskConfiguration:
    """Annotate a DebusineTaskConfiguration with path information."""
    return sample.model_copy(update={"path": path, "position": position})


def simplify_dicts(values: list[dict[str, Any]]) -> list[dict[str, Any]]:
    """Given a list of dicts, returns the same with the common keys removed."""
    assert values

    # Start with a dict with the contents of the first dict
    same = values[0].copy()

    # Drop all elements that change in other dicts
    for value in values[1:]:
        for k, v in value.items():
            if same.get(k) != v:
                same.pop(k, None)

    result: list[dict[str, Any]] = []
    for value in values:
        result.append({k: v for k, v in value.items() if k not in same})  # noqa: PERF401
    return result


class TaskConfigurationTestCase(TestCase):
    """Common function for task configuration related tests."""

    logger: ClassVar[logging.Logger]

    @override
    @classmethod
    def setUpClass(cls) -> None:
        super().setUpClass()
        cls.logger = logging.getLogger("debusine.client.tests")

    @override
    def setUp(self) -> None:
        super().setUp()
        self.root = self.create_temporary_directory()
        self.samples = Samples()

    def manifest(
        self,
        workspace: str = "System",
        pk: int = 42,
        name: str = "default",
        data: dict[str, Any] | None = None,
    ) -> Manifest:
        """Create a Manifest."""
        return Manifest(
            workspace=workspace,
            collection=TaskConfigurationCollection(
                id=pk, name=name, data=data or {}
            ),
        )

    def make_remote_repo(
        self, entries: list[DebusineTaskConfiguration] | None = None
    ) -> RemoteTaskConfigurationRepository:
        repo = RemoteTaskConfigurationRepository(self.manifest())
        repo.entries.extend(entries or [])
        return repo

    def write(
        self,
        relpath: Path,
        item: pydantic.BaseModel | Collection[pydantic.BaseModel],
    ) -> None:
        """Write an item to a file in self.root."""
        path = self.root / relpath
        path.parent.mkdir(exist_ok=True, parents=True)
        with path.open("wt") as fd:
            if isinstance(item, pydantic.BaseModel):
                yaml.safe_dump(item.model_dump(mode="json"), fd)
            else:
                yaml.safe_dump([i.model_dump(mode="json") for i in item], fd)

    def assertFileContents(
        self,
        relpath: Path,
        expected: pydantic.BaseModel | Collection[pydantic.BaseModel],
    ) -> None:
        """Check that path points to a YAML file with the given contents."""
        path = self.root / relpath
        with path.open() as fd:
            actual = yaml.safe_load(fd)
        if isinstance(expected, pydantic.BaseModel):
            assert actual == expected.model_dump(mode="json")
        else:
            assert simplify_dicts(actual) == simplify_dicts(
                [i.model_dump(mode="json") for i in expected]
            )

    def assertPullLog(self, log: _LoggingWatcher, lines: list[str]) -> None:
        """Check that the pull log contains the expected lines."""
        self.assertCountEqual(
            [x.removeprefix("INFO:debusine.client.tests:") for x in log.output],
            lines,
        )

    def list_files(self) -> list[Path]:
        """List files present in self.root."""
        res: list[Path] = []
        for cur_root, _dirs, files in self.root.walk():
            res += [(cur_root / name).relative_to(self.root) for name in files]
        return sorted(res)


class PullStatsTests(TestCase):
    """Test PullStats class."""

    def test_has_changes(self) -> None:
        for args, expected in (
            ((0, 0, 0, 0), False),
            ((1, 0, 0, 0), True),
            ((0, 1, 0, 0), True),
            ((0, 0, 1, 0), True),
            ((0, 0, 0, 1), False),
            ((1, 1, 1, 1), True),
        ):
            with self.subTest(args=args):
                stats = PullStats(*args)
                self.assertEqual(stats.has_changes(), expected)

    def test_eq(self) -> None:
        for first, second, expected in (
            (PullStats(), PullStats(), True),
            (PullStats(1), PullStats(1), True),
            (PullStats(1), PullStats(2), False),
            (PullStats(1, 1), PullStats(1, 2), False),
            (PullStats(1, 1, 1), PullStats(1, 1, 2), False),
            (PullStats(1, 1, 1, 1), PullStats(1, 1, 1, 2), False),
            (PullStats(1), 1, False),
        ):
            with self.subTest(first=repr(first), second=repr(second)):
                if expected:
                    self.assertEqual(first, second)
                else:
                    self.assertNotEqual(first, second)


class EntryFileTests(TaskConfigurationTestCase):
    """Tests for :py:class:`EntryFile`."""

    @override
    def setUp(self) -> None:
        super().setUp()
        self.relpath = Path("test.yaml")
        self.ef = EntryFile(self.relpath)

    def test_empty(self) -> None:
        self.assertEqual(self.ef.relpath, self.relpath)
        self.assertEqual(self.ef.entries, [])

    def test_load_single(self) -> None:
        sample = DebusineTaskConfiguration(template="test")
        self.write(self.relpath, sample)
        self.ef.load(self.root)
        self.assertEqual(self.ef.entries, items(sample))

    def test_load_multi(self) -> None:
        self.write(self.relpath, [self.samples.s1, self.samples.s2])
        self.ef.load(self.root)
        self.assertEqual(
            self.ef.entries, items(self.samples.s1, self.samples.s2)
        )

    def test_load_annotates_exceptions(self) -> None:
        with self.assertRaises(FileNotFoundError) as e:
            self.ef.load(self.root)
        self.assertEqual(e.exception.__notes__, ["source: test.yaml"])

    def test_load_strips_path_position(self) -> None:
        sample = DebusineTaskConfiguration(
            template="test1", path="foo.yaml", position=3
        )
        self.write(self.relpath, sample)
        self.ef.load(self.root)
        self.assertEqual(self.ef.entries, [Item(template="test1")])

    def test_write_single(self) -> None:
        self.ef.entries += items(self.samples.s)
        self.ef.write(self.root)
        self.assertFileContents(self.relpath, self.samples.s)

    def test_write_multi(self) -> None:
        self.ef.entries += items(self.samples.s2, self.samples.s1)
        self.ef.write(self.root)
        self.assertFileContents(
            self.relpath, [self.samples.s2, self.samples.s1]
        )

    def test_write_unchanged(self) -> None:
        sample = Item(template="test")
        self.write(self.relpath, sample)
        os.utime(self.root / self.relpath, (123456, 123456))

        self.ef.entries.append(sample)
        self.ef.write(self.root)
        st = (self.root / self.relpath).stat()
        self.assertEqual(st.st_mtime, 123456)

    def test_diff(self) -> None:
        ef1 = EntryFile(Path("file1"))
        ef1.entries += items(self.samples.s1, self.samples.s2, self.samples.s3)
        ef2 = EntryFile(Path("file2"))
        ef2.entries += items(
            self.samples.s1,
            self.samples.s2,
            self.samples.s4,
            self.samples.s5,
            self.samples.s6,
        )
        stats = ef1.diff(ef2)
        self.assertEqual(stats.added, 2)
        self.assertEqual(stats.deleted, 0)
        self.assertEqual(stats.updated, 1)
        self.assertEqual(stats.unchanged, 2)

    def test_diff_deleted(self) -> None:
        ef1 = EntryFile(Path("file1"))
        ef1.entries += items(self.samples.s1, self.samples.s2, self.samples.s3)
        ef2 = EntryFile(Path("file2"))
        ef2.entries += items(self.samples.s1)
        stats = ef1.diff(ef2)
        self.assertEqual(stats.added, 0)
        self.assertEqual(stats.deleted, 2)
        self.assertEqual(stats.updated, 0)
        self.assertEqual(stats.unchanged, 1)

    def test_diff_replace_with_shorter_sequence(self) -> None:
        ef1 = EntryFile(Path("file1"))
        ef1.entries += items(
            self.samples.s1, self.samples.s2, self.samples.s3, self.samples.s4
        )
        ef2 = EntryFile(Path("file2"))
        ef2.entries += items(self.samples.s1, self.samples.s5, self.samples.s4)
        stats = ef1.diff(ef2)
        self.assertEqual(stats.added, 0)
        self.assertEqual(stats.deleted, 1)
        self.assertEqual(stats.updated, 1)
        self.assertEqual(stats.unchanged, 2)


class RemoteTaskConfigurationRepositoryTestsBase(TaskConfigurationTestCase):
    """Tests for :py:class:`RemoteTaskConfigurationRepository`."""

    def test_from_items(self) -> None:
        repo = RemoteTaskConfigurationRepository.from_items(
            self.manifest(), [self.samples.s1, self.samples.s2, self.samples.s3]
        )
        self.assertEqual(repo.manifest, self.manifest())
        self.assertEqual(
            repo.entries, [self.samples.s1, self.samples.s2, self.samples.s3]
        )


class LocalTaskConfigurationRepositoryTestCase(TaskConfigurationTestCase):
    """Base for LocalTaskConfigurationRepository tests."""

    @override
    def setUp(self) -> None:
        super().setUp()
        self.path_manifest = Path("MANIFEST")

    def populate(
        self,
        manifest: Manifest | None = None,
        files: (
            dict[
                str | Path,
                DebusineTaskConfiguration | list[DebusineTaskConfiguration],
            ]
            | None
        ) = None,
    ) -> None:
        """Populate the local repository."""
        self.write(self.path_manifest, manifest or self.manifest())
        for path, entries in (files or {}).items():
            if isinstance(entries, DebusineTaskConfiguration):
                self.write(Path(path), as_item(entries))
            else:
                self.write(Path(path), items(*entries))


class LocalTaskConfigurationRepositoryTests(
    LocalTaskConfigurationRepositoryTestCase
):
    """Test LocalTaskConfigurationRepository class."""

    def make_local_repo(
        self,
    ) -> LocalTaskConfigurationRepository:
        """Create a LocalTaskConfigurationRepository with manifest."""
        return LocalTaskConfigurationRepository(self.root)

    def test_is_dirty(self) -> None:
        repo = self.make_local_repo()
        self.assertFalse(repo.is_dirty())

    def test_git_commit(self) -> None:
        repo = self.make_local_repo()
        self.assertIsNone(repo.git_commit())

    def test_has_commit(self) -> None:
        repo = self.make_local_repo()
        self.assertFalse(repo.has_commit("0" * 40))

    def test_save_manifest_annotates_exceptions(self) -> None:
        repo = self.make_local_repo()
        repo.manifest = self.manifest()
        (repo.root / "MANIFEST").mkdir()
        with self.assertRaises(OSError) as exc:
            repo.write_manifest()
        self.assertEqual(
            exc.exception.__notes__, [f"manifest: {repo.root}/MANIFEST"]
        )

    def test_init_empty_dir(self) -> None:
        repo = self.make_local_repo()
        self.assertEqual(repo.root, self.root)
        self.assertEqual(repo.manifest_path, self.root / "MANIFEST")
        self.assertIsNone(repo.manifest)
        self.assertEqual(repo.files, {})

    def test_init(self) -> None:
        self.populate(
            files={
                (path1 := Path("path1.yaml")): [self.samples.s1],
                (path2 := Path("path2.yaml")): [
                    self.samples.s2,
                    self.samples.s3,
                ],
            }
        )
        repo = self.make_local_repo()
        self.assertEqual(repo.root, self.root)
        self.assertEqual(repo.manifest_path, self.root / "MANIFEST")
        self.assertEqual(repo.manifest, self.manifest())
        self.assertCountEqual(repo.files.keys(), {path1, path2})
        self.assertEqual(repo.files[path1].entries, items(self.samples.s1))
        self.assertEqual(
            repo.files[path2].entries, items(self.samples.s2, self.samples.s3)
        )

    def test_iter_files(self) -> None:
        self.write(path1 := Path("path1.yaml"), [])
        self.write(path2 := Path("path2.yaml"), [])
        self.write(path3 := Path("path3/path.yaml"), [])
        self.assertEqual(
            set(LocalTaskConfigurationRepository.iter_files(self.root)),
            {path1, path2, path3},
        )

    def test_iter_files_ignore_non_yaml(self) -> None:
        (self.root / "README").write_text("Thank you for reading")
        (self.root / "data.json").write_text("{}")
        self.assertEqual(
            list(LocalTaskConfigurationRepository.iter_files(self.root)), []
        )

    def test_iter_entries(self) -> None:
        self.populate(
            files={
                "single.yaml": self.samples.s1,
                "multi.yaml": [self.samples.s2, self.samples.s3],
                "dup.yaml": [self.samples.s1, self.samples.s1],
            },
        )
        local_repo = self.make_local_repo()

        self.assertCountEqual(
            list(local_repo.iter_entries()),
            [
                with_path(self.samples.s1, "single.yaml", 0),
                with_path(self.samples.s2, "multi.yaml", 0),
                with_path(self.samples.s3, "multi.yaml", 1),
                with_path(self.samples.s1, "dup.yaml", 0),
                with_path(self.samples.s1, "dup.yaml", 1),
            ],
        )

    def test_write_manifest_unset(self) -> None:
        repo = self.make_local_repo()
        with self.assertRaisesRegex(
            InvalidRepository, "Repository manifest has not been set"
        ):
            repo.write_manifest()

    def test_write_manifest(self) -> None:
        repo = self.make_local_repo()
        repo.manifest = self.manifest()
        self.assertFalse(repo.manifest_path.exists())
        repo.write_manifest()
        self.assertTrue(repo.manifest_path.exists())
        self.assertFileContents(Path("MANIFEST"), repo.manifest)

    def test_write_entries(self) -> None:
        local_repo = self.make_local_repo()
        ef = EntryFile(Path("test.yaml"))
        local_repo.files[ef.relpath] = ef
        ef.entries = items(self.samples.s1, self.samples.s2)
        local_repo.write_entries()
        self.assertFileContents(ef.relpath, [self.samples.s1, self.samples.s2])

    def test_write_entries_removes_unknown(self) -> None:
        self.populate(
            files={
                (path_removed := Path("removed.yaml")): self.samples.s1,
                (path_kept := Path("kept.yaml")): self.samples.s2,
            }
        )
        local_repo = self.make_local_repo()
        del local_repo.files[path_removed]
        local_repo.write_entries()
        self.assertCountEqual(
            self.list_files(), [self.path_manifest, path_kept]
        )
        self.assertFileContents(path_kept, self.samples.s2)

    def test_pull_new_repo(self) -> None:
        server_repo = self.make_remote_repo([self.samples.s1, self.samples.s2])
        local_repo = self.make_local_repo()
        with self.assertLogs(self.logger) as log:
            stats = local_repo.pull(server_repo, self.logger)
        self.assertPullLog(log, ["new.yaml: new file with 2 entries"])
        self.assertEqual(stats, PullStats(added=2))
        self.assertCountEqual(
            self.list_files(),
            [
                path_manifest := Path("MANIFEST"),
                path_new := Path("new.yaml"),
            ],
        )
        self.assertFileContents(path_manifest, server_repo.manifest)
        self.assertFileContents(path_new, [self.samples.s1, self.samples.s2])

    def test_pull_new_repo_with_paths(self) -> None:
        self.samples.add_config("root", path="test.yaml")
        self.samples.add_config(
            "foo1", path="foo/test.yaml", position=1, subject="s1"
        )
        self.samples.add_config(
            "foo2", path="foo/test.yaml", position=2, subject="s2"
        )
        server_repo = self.make_remote_repo(
            [self.samples.root, self.samples.foo1, self.samples.foo2]
        )
        local_repo = self.make_local_repo()
        with self.assertLogs(self.logger) as log:
            stats = local_repo.pull(server_repo, self.logger)
        self.assertPullLog(
            log,
            [
                "test.yaml: new file with 1 entries",
                "foo/test.yaml: new file with 2 entries",
            ],
        )
        self.assertEqual(stats, PullStats(added=3))
        self.assertCountEqual(
            self.list_files(),
            [
                self.path_manifest,
                path_root := Path("test.yaml"),
                path_foo := Path("foo/test.yaml"),
            ],
        )
        self.assertFileContents(self.path_manifest, server_repo.manifest)
        self.assertFileContents(path_root, as_item(self.samples.root))
        self.assertFileContents(
            path_foo, items(self.samples.foo1, self.samples.foo2)
        )

    def test_pull_no_remote_paths(self) -> None:
        server_repo = self.make_remote_repo([self.samples.s1, self.samples.s2])

        path_samples = Path("samples.yaml")
        self.populate(
            server_repo.manifest,
            files={path_samples: [self.samples.s1, self.samples.s2]},
        )
        local_repo = self.make_local_repo()

        with self.assertNoLogs(self.logger.name):
            stats = local_repo.pull(server_repo, self.logger)
        self.assertEqual(stats, PullStats(unchanged=2))
        self.assertCountEqual(
            self.list_files(),
            [self.path_manifest, path_samples],
        )
        self.assertFileContents(self.path_manifest, server_repo.manifest)
        self.assertFileContents(
            path_samples, items(self.samples.s1, self.samples.s2)
        )

    def test_pull_append_pathless_to_existing_new_yaml(self) -> None:
        server_repo = self.make_remote_repo(
            [
                self.samples.add_config("p1", path="new.yaml", position=0),
                self.samples.add_config("p2", path="new.yaml", position=2),
                self.samples.s1,
                self.samples.s2,
            ]
        )

        path_manifest = Path("MANIFEST")
        path_new = Path("new.yaml")

        self.populate(
            manifest=server_repo.manifest, files={path_new: [self.samples.s1]}
        )
        local_repo = self.make_local_repo()

        with self.assertLogs(self.logger.name) as log:
            stats = local_repo.pull(server_repo, self.logger)
        self.assertPullLog(
            log, ["new.yaml: 3 added, 0 updated, 0 deleted, 1 unchanged"]
        )
        self.assertEqual(stats, PullStats(added=3, unchanged=1))
        self.assertCountEqual(
            self.list_files(),
            [path_manifest, path_new],
        )
        self.assertFileContents(path_manifest, server_repo.manifest)
        self.assertFileContents(
            path_new,
            items(
                self.samples.p1,
                self.samples.s1,
                self.samples.p2,
                self.samples.s2,
            ),
        )

    def test_pull_update(self) -> None:
        old = self.samples.add_config("s1", path="test.yaml", position=1)
        new = old.model_copy(update={"comment": "updated"})

        server_repo = self.make_remote_repo([new])

        assert old.path is not None
        path_test = Path(old.path)
        self.populate(manifest=server_repo.manifest, files={path_test: [old]})
        local_repo = self.make_local_repo()

        with self.assertLogs(self.logger) as log:
            stats = local_repo.pull(server_repo, self.logger)
        self.assertPullLog(
            log, ["test.yaml: 0 added, 1 updated, 0 deleted, 0 unchanged"]
        )
        self.assertEqual(stats, PullStats(updated=1))
        self.assertCountEqual(
            self.list_files(), [self.path_manifest, path_test]
        )
        self.assertFileContents(self.path_manifest, server_repo.manifest)
        self.assertFileContents(path_test, as_item(new))

    def test_pull_update_preserves_mtime(self) -> None:
        self.samples.add_config("changed", path="changed.yaml", position=1)
        new = self.samples.changed.model_copy(update={"comment": "updated"})

        server_repo = self.make_remote_repo(
            [new, self.samples.add_config("unchanged", path="unchanged.yaml")]
        )

        path_changed = Path("changed.yaml")
        path_unchanged = Path("unchanged.yaml")
        self.populate(
            manifest=server_repo.manifest,
            files={
                path_changed: self.samples.changed,
                path_unchanged: self.samples.unchanged,
            },
        )
        oldtime = tm.time_ns() - 3600 * 1_000_000_000
        os.utime(self.root / path_changed, ns=(oldtime, oldtime))
        os.utime(self.root / path_unchanged, ns=(oldtime, oldtime))
        orig_stat_changed = (self.root / path_changed).stat()
        orig_stat_unchanged = (self.root / path_unchanged).stat()

        local_repo = self.make_local_repo()

        with self.assertLogs(self.logger) as log:
            stats = local_repo.pull(server_repo, self.logger)
        self.assertPullLog(
            log, ["changed.yaml: 0 added, 1 updated, 0 deleted, 0 unchanged"]
        )
        self.assertEqual(stats, PullStats(updated=1, unchanged=1))
        self.assertCountEqual(
            self.list_files(),
            [self.path_manifest, path_changed, path_unchanged],
        )
        self.assertFileContents(self.path_manifest, server_repo.manifest)
        self.assertFileContents(path_changed, as_item(new))
        self.assertFileContents(path_unchanged, as_item(self.samples.unchanged))
        self.assertLess(
            orig_stat_changed.st_mtime_ns,
            (self.root / path_changed).stat().st_mtime_ns,
        )
        self.assertEqual(
            orig_stat_unchanged.st_mtime_ns,
            (self.root / path_unchanged).stat().st_mtime_ns,
        )

    def test_pull_update_manifest_collection_data(self) -> None:
        server_repo = self.make_remote_repo()
        server_repo.manifest = self.manifest(data={"test": True})

        path_manifest = Path("MANIFEST")
        self.write(path_manifest, self.manifest())
        local_repo = self.make_local_repo()

        with self.assertNoLogs(self.logger.name):
            stats = local_repo.pull(server_repo, self.logger)
        self.assertEqual(stats, PullStats())
        self.assertCountEqual(self.list_files(), [path_manifest])
        self.assertFileContents(path_manifest, server_repo.manifest)

    def test_pull_delete(self) -> None:
        server_repo = self.make_remote_repo()

        self.populate(
            manifest=server_repo.manifest,
            files={Path("path1.yaml"): self.samples.s1},
        )
        local_repo = self.make_local_repo()

        with self.assertLogs(self.logger) as log:
            stats = local_repo._pull(
                server_repo, self.logger, preserve_deleted=False
            )
        self.assertPullLog(log, ["path1.yaml: file with 1 entries deleted"])
        self.assertEqual(stats, PullStats(deleted=1))
        self.assertCountEqual(self.list_files(), [self.path_manifest])
        self.assertFileContents(self.path_manifest, server_repo.manifest)

    def test_pull_delete_preserve(self) -> None:
        server_repo = self.make_remote_repo()

        self.populate(
            manifest=server_repo.manifest,
            files={(path1 := Path("path1.yaml")): self.samples.s1},
        )
        local_repo = self.make_local_repo()

        with self.assertLogs(self.logger) as log:
            stats = local_repo._pull(
                server_repo, self.logger, preserve_deleted=True
            )
        self.assertPullLog(
            log, ["path1.yaml: file with 1 entries preserved from deletion"]
        )
        self.assertEqual(stats, PullStats(deleted=1))
        self.assertCountEqual(self.list_files(), [self.path_manifest, path1])
        self.assertFileContents(self.path_manifest, server_repo.manifest)
        self.assertFileContents(path1, self.samples.s1)

    def test_pull_delete_in_bundle(self) -> None:
        server_repo = self.make_remote_repo([self.samples.s2])

        self.populate(
            manifest=server_repo.manifest,
            files={
                (path1 := Path("path1.yaml")): [
                    self.samples.s1,
                    self.samples.s2,
                ]
            },
        )
        local_repo = self.make_local_repo()

        with self.assertLogs(self.logger) as log:
            stats = local_repo._pull(
                server_repo, self.logger, preserve_deleted=False
            )
        self.assertPullLog(
            log, ["path1.yaml: 0 added, 0 updated, 1 deleted, 1 unchanged"]
        )
        self.assertEqual(stats, PullStats(deleted=1, unchanged=1))
        self.assertCountEqual(
            self.list_files(),
            [self.path_manifest, path1],
        )
        self.assertFileContents(self.path_manifest, server_repo.manifest)
        self.assertFileContents(path1, as_item(self.samples.s2))

    def test_pull_delete_clean_empty_dirs(self) -> None:
        self.samples.add_config("keep", subject="keep")
        server_repo = self.make_remote_repo([self.samples.keep])

        self.populate(
            manifest=server_repo.manifest,
            files={
                Path("dir1/file.yaml"): self.samples.s1,
                (path_keep := Path("dir1/dir1/file.yaml")): self.samples.keep,
                Path("dir1/dir2/file.yaml"): self.samples.s2,
                Path("dir2/dir1/file.yaml"): self.samples.s3,
                Path("dir2/dir2/file.yaml"): self.samples.s4,
            },
        )
        local_repo = self.make_local_repo()
        with self.assertLogs(self.logger) as log:
            stats = local_repo._pull(
                server_repo, self.logger, preserve_deleted=False
            )
        self.assertPullLog(
            log,
            [
                "dir1/file.yaml: file with 1 entries deleted",
                "dir1/dir2/file.yaml: file with 1 entries deleted",
                "dir2/dir1/file.yaml: file with 1 entries deleted",
                "dir2/dir2/file.yaml: file with 1 entries deleted",
            ],
        )
        self.assertEqual(stats, PullStats(deleted=4, unchanged=1))
        self.assertCountEqual(
            self.list_files(), [self.path_manifest, path_keep]
        )
        self.assertFileContents(self.path_manifest, server_repo.manifest)
        self.assertTrue((self.root / "dir1").exists())
        self.assertTrue((self.root / "dir1/dir1").exists())
        self.assertFalse((self.root / "dir1/dir2").exists())
        self.assertFalse((self.root / "dir2").exists())

    def test_preserve_all_matching_pathless_entries(self) -> None:
        self.samples.add_config("keep", subject="keep")
        server_repo = self.make_remote_repo([self.samples.keep])

        self.populate(
            manifest=server_repo.manifest,
            files={
                Path("path1.yaml"): self.samples.s1,
                Path("path2.yaml"): self.samples.s2,
                (path_keep1 := Path("keep1.yaml")): [
                    self.samples.keep,
                    self.samples.s3,
                ],
                (path_keep2 := Path("keep2.yaml")): [
                    self.samples.s4,
                    self.samples.keep,
                    self.samples.s5,
                ],
            },
        )
        local_repo = self.make_local_repo()
        with self.assertLogs(self.logger) as log:
            stats = local_repo._pull(
                server_repo, self.logger, preserve_deleted=False
            )
        self.assertPullLog(
            log,
            [
                "path1.yaml: file with 1 entries deleted",
                "path2.yaml: file with 1 entries deleted",
                "keep1.yaml: 0 added, 0 updated, 1 deleted, 1 unchanged",
                "keep2.yaml: 0 added, 0 updated, 2 deleted, 1 unchanged",
            ],
        )
        self.assertEqual(stats, PullStats(deleted=5, unchanged=2))
        self.assertCountEqual(
            self.list_files(), [self.path_manifest, path_keep1, path_keep2]
        )
        self.assertFileContents(self.path_manifest, server_repo.manifest)
        self.assertFileContents(path_keep1, as_item(self.samples.keep))
        self.assertFileContents(path_keep2, as_item(self.samples.keep))

    def test_pull_repo_manifest_mismatch(self) -> None:
        server_repo = self.make_remote_repo()

        path_manifest = Path("MANIFEST")
        self.write(path_manifest, self.manifest(pk=7))
        local_repo = self.make_local_repo()

        with self.assertRaisesRegex(
            InvalidRepository,
            r"repo to pull refers to collection System/default \(42\)"
            r" while the checkout has collection System/default \(7\)",
        ):
            local_repo.pull(server_repo, self.logger)

        self.assertCountEqual(self.list_files(), [Path("MANIFEST")])

    def test_is_checkout(self) -> None:
        self.assertFalse(
            LocalTaskConfigurationRepository.is_checkout(self.root)
        )
        (self.root / "MANIFEST").write_text("")
        self.assertTrue(LocalTaskConfigurationRepository.is_checkout(self.root))

    def test_read_manifest_annotates_exceptions(self) -> None:
        with self.assertRaises(OSError) as exc:
            LocalTaskConfigurationRepository.read_manifest(self.root)
        self.assertEqual(
            exc.exception.__notes__, [f"manifest: {self.root}/MANIFEST"]
        )


class GitTaskConfigurationRepositoryTests(
    LocalTaskConfigurationRepositoryTestCase
):
    """Test GitTaskConfigurationRepository class."""

    @override
    @classmethod
    def setUpClass(cls) -> None:
        super().setUpClass()
        if shutil.which("git") is None:  # pragma: no cover
            raise SkipTest("git is not installed")

    @override
    def setUp(self) -> None:
        super().setUp()
        self.git("init", "-b", "main")
        self.git("config", "user.name", "Test User")
        self.git("config", "user.email", "test@example.com")

    def make_local_repo(
        self,
    ) -> GitTaskConfigurationRepository:
        """Create a LocalTaskConfigurationRepository with manifest."""
        return GitTaskConfigurationRepository(self.root)

    def git(self, *args: str) -> subprocess.CompletedProcess[str]:
        """Run a git command in the repository and return its output."""
        cmd = ["git", "-C", self.root.as_posix()]
        cmd.extend(args)
        return subprocess.run(
            cmd,
            cwd=self.root,
            capture_output=True,
            text=True,
            check=True,
            env={
                "GIT_CONFIG_GLOBAL": "/dev/null",
                "GIT_CONFIG_SYSTEM": "/dev/null",
            },
        )

    def commit(self, message: str = "Commit message") -> None:
        """Commit changes in self.root."""
        self.git("add", ".")
        self.git("commit", "-m", message)

    def test_from_path(self) -> None:
        repo1 = LocalTaskConfigurationRepository.from_path(self.root)
        self.assertIsInstance(repo1, GitTaskConfigurationRepository)
        repo2 = LocalTaskConfigurationRepository.from_path(
            self.create_temporary_directory()
        )
        self.assertIsInstance(repo2, LocalTaskConfigurationRepository)

    def test_git_error_annotates_exception(self) -> None:
        repo = self.make_local_repo()
        shutil.rmtree(self.root / ".git")
        with self.assertRaises(subprocess.CalledProcessError) as exc:
            repo.git(["rev-parse", "HEAD"], check=True)
        self.assertEqual(exc.exception.__notes__[0][:8], "stderr:\n")

    def test_is_git(self) -> None:
        self.assertTrue(GitTaskConfigurationRepository.is_git(self.root))
        self.assertFalse(
            GitTaskConfigurationRepository.is_git(
                self.create_temporary_directory()
            )
        )
        with mock.patch("shutil.which", return_value=None):
            self.assertFalse(GitTaskConfigurationRepository.is_git(self.root))

    def test_is_dirty_empty(self) -> None:
        repo = self.make_local_repo()
        self.assertFalse(repo.is_dirty())

    def test_is_dirty_uncommitted_manifest(self) -> None:
        repo = self.make_local_repo()
        repo.manifest_path.write_text("test")
        self.assertTrue(repo.is_dirty())

    def test_is_dirty_uncommitted_yaml(self) -> None:
        repo = self.make_local_repo()
        self.write(Path("config.yaml"), self.samples.s1)
        self.assertTrue(repo.is_dirty())

    def test_is_dirty_dir(self) -> None:
        repo = self.make_local_repo()
        (self.root / "testdir").mkdir()
        (self.root / "testdir" / "testfile").write_text("")
        self.assertTrue(repo.is_dirty())

    def test_is_dirty_uncommitted_other(self) -> None:
        repo = self.make_local_repo()
        self.write(Path("README"), self.samples.s1)
        self.assertFalse(repo.is_dirty())

    def test_git_commit_empty(self) -> None:
        repo = self.make_local_repo()
        self.assertIsNone(repo.git_commit())

    def test_git_commit(self) -> None:
        repo = self.make_local_repo()
        repo.manifest = self.manifest()
        repo.write_manifest()
        self.commit()
        self.assertFalse(repo.is_dirty())
        self.assertIsNotNone(repo.git_commit())

    def test_has_commit(self) -> None:
        repo = self.make_local_repo()
        repo.manifest = self.manifest()
        repo.write_manifest()
        self.commit()
        first_commit = repo.git_commit()
        assert first_commit is not None
        self.write(Path("sample1.yaml"), self.samples.s1)
        self.commit()
        second_commit = repo.git_commit()
        assert second_commit is not None
        self.write(Path("sample2.yaml"), self.samples.s2)
        self.commit()
        third_commit = repo.git_commit()
        assert third_commit is not None
        self.assertTrue(repo.has_commit(first_commit))
        self.assertTrue(repo.has_commit(second_commit))
        self.assertTrue(repo.has_commit(third_commit))

        missing_commit = (
            ("0" if first_commit[0] != "0" else "1")
            + ("0" if second_commit[1] != "0" else "1")
            + ("0" if third_commit[2] != "0" else "1")
            + first_commit[3:]
        )
        self.assertFalse(repo.has_commit(missing_commit))

    def test_pull_dirty(self) -> None:
        server_repo = self.make_remote_repo()
        self.write(Path("config.yaml"), self.samples.s1)
        local_repo = self.make_local_repo()
        with self.assertRaisesRegex(
            InvalidRepository,
            r"git working directory is dirty: please commit changes"
            r" before pulling from debusine",
        ):
            local_repo.pull(server_repo, self.logger)

    def test_pull(self) -> None:
        commit = "0" * 40
        server_repo = self.make_remote_repo()
        server_repo.manifest.collection.data["git_commit"] = commit
        self.write(Path("MANIFEST"), self.manifest(data={"git_commit": commit}))
        self.commit()
        local_repo = self.make_local_repo()
        with self.assertNoLogs(self.logger.name):
            local_repo.pull(server_repo, self.logger)



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Debusine command line interface, collection relation commands."""

import argparse
import sys
from collections.abc import Iterable
from operator import attrgetter
from typing import override

import rich
from rich.table import Table

from debusine.artifacts.models import CollectionCategory
from debusine.client.commands.base import ModelCommand, WorkspaceCommand
from debusine.client.exceptions import DebusineError
from debusine.client.models import (
    CollectionIdentifierRequest,
    CollectionRelationResponse,
    CollectionRelationSetResponse,
    CollectionRelationType,
)
from debusine.utils.input import YamlEditor


def _validate_collection_identifier(value: str) -> None:
    """Raise ValueError if the format or category are invalid."""
    invalid_collection_identifier_error = (
        f"Invalid collection identifier: {value!r}. "
        "Expected format is 'name@category'"
    )

    if "@" not in value:
        raise ValueError(invalid_collection_identifier_error)

    name, category = value.split("@", maxsplit=1)

    if name == "" or category == "":
        raise ValueError(invalid_collection_identifier_error)

    valid_categories = [
        item
        for item in CollectionCategory
        if item is not CollectionCategory.TEST
    ]

    if category not in valid_categories:
        categories = ", ".join(item.value for item in sorted(valid_categories))

        raise ValueError(
            f"Invalid collection identifier: {value!r}. "
            f"Category must be one of: {categories}"
        )


def _collection_identifier_argument(value: str) -> str:
    """Validate a collection string. Raise ArgumentTypeError."""
    try:
        _validate_collection_identifier(value)
    except ValueError as exc:
        raise argparse.ArgumentTypeError(str(exc)) from exc

    return value


class List(
    ModelCommand[CollectionRelationResponse],
    WorkspaceCommand,
    group="collection.relation",
):
    """List relations between collections in a workspace."""

    @override
    @classmethod
    def configure(cls, parser: argparse.ArgumentParser) -> None:
        super().configure(parser)
        parser.add_argument(
            "--from",
            dest="source",
            type=_collection_identifier_argument,
            metavar="COLLECTION",
            help="Filter by source collection in name@category format",
        )
        parser.add_argument(
            "--to",
            dest="target",
            type=_collection_identifier_argument,
            metavar="COLLECTION",
            help="Filter by target collection in name@category format",
        )
        parser.add_argument(
            "--type",
            dest="relation_type",
            choices=sorted(CollectionRelationType, key=attrgetter("value")),
            metavar="TYPE",
            help="Filter by relation type",
        )

    @override
    def run(self) -> None:
        with self._api_call_or_fail():
            relations = self.debusine.collection_relation_list(
                self.workspace,
                source=self.args.source,
                target=self.args.target,
                relation_type=self.args.relation_type,
            )
        self.list(relations)

    @override
    def _list_rich(
        self, collection_relations: Iterable[CollectionRelationResponse]
    ) -> None:
        """Print the list as a table."""
        table = Table(box=rich.box.MINIMAL_DOUBLE_HEAD)
        table.add_column("From")
        table.add_column("To")
        table.add_column("Type")
        table.add_column("Position")

        for relation in collection_relations:
            table.add_row(
                (
                    f"{relation.source_collection.name}"
                    f"@{relation.source_collection.category}"
                ),
                (
                    f"{relation.target_collection.name}"
                    f"@{relation.target_collection.category}"
                ),
                str(relation.type),
                (
                    str(relation.position)
                    if relation.position is not None
                    else "-"
                ),
            )
        rich.print(table)

    @override
    def _show_rich(
        self, collection_relation: CollectionRelationResponse
    ) -> None:
        raise NotImplementedError()


def _collection_identifiers_to_strs(
    collection_relations: Iterable[CollectionRelationResponse],
) -> list[str]:
    """Return target collection identifiers as strings."""
    return [
        f"{relation.target_collection.name}@"
        f"{relation.target_collection.category}"
        for relation in collection_relations
    ]


def _collection_identifiers_strs_to_requests(
    targets: list[str],
) -> list[CollectionIdentifierRequest]:
    requests = []

    for target in targets:
        _validate_collection_identifier(target)
        name, category = target.split("@", maxsplit=1)
        requests.append(
            CollectionIdentifierRequest(
                name=name, category=CollectionCategory(category)
            )
        )

    return requests


class Edit(WorkspaceCommand, group="collection.relation"):
    """Edit the relation target list for a given source collection and type."""

    @override
    @classmethod
    def configure(cls, parser: argparse.ArgumentParser) -> None:
        super().configure(parser)
        parser.add_argument(
            "source",
            type=_collection_identifier_argument,
            metavar="FROM-NAME@CATEGORY",
            help="Source collection in name@category format",
        )
        parser.add_argument(
            "relation_type",
            choices=sorted(CollectionRelationType, key=attrgetter("value")),
            metavar="TYPE",
            help="Relation type to edit",
        )
        parser.add_argument(
            "--append",
            type=_collection_identifier_argument,
            nargs="+",
            metavar="COLLECTION",
            help="Add collection(s) to the end of the target list",
        )
        parser.add_argument(
            "--prepend",
            type=_collection_identifier_argument,
            nargs="+",
            metavar="COLLECTION",
            help="Add collection(s) to the beginning of the target list",
        )
        parser.add_argument(
            "--remove",
            type=_collection_identifier_argument,
            nargs="+",
            metavar="COLLECTION",
            help="Remove collection(s) from the target list",
        )
        parser.add_argument(
            "--set",
            type=_collection_identifier_argument,
            nargs="+",
            metavar="COLLECTION",
            help="Replace the complete target list with collection(s)",
        )
        parser.add_argument(
            "--edit",
            action="store_true",
            help="Edit the complete target list using $EDITOR",
        )

    def _apply_target_list(
        self,
        *,
        source: str,
        relation_type: CollectionRelationType,
        targets: list[str],
        if_match: str,
    ) -> CollectionRelationSetResponse:
        """Apply a complete target list for (source, relation_type)."""
        request_targets = _collection_identifiers_strs_to_requests(targets)
        with self._api_call_or_fail():
            return self.debusine.collection_relation_set(
                self.workspace,
                source=source,
                relation_type=relation_type,
                targets=request_targets,
                if_match=if_match,
            )

    def _edit_and_apply_targets(
        self,
        source: str,
        relation_type: CollectionRelationType,
        initial: list[str],
        if_match: str,
    ) -> CollectionRelationSetResponse | None:
        """Run the editor and apply new targets."""
        editor: YamlEditor[list[str]] = YamlEditor(
            initial,
            help_text=[
                "Here you can edit the relation target list.",
                "",
                "This is a YAML list of collection identifiers",
                "written as NAME@CATEGORY.",
                "Each entry represents a target collection",
                f"from {source} and relation type {relation_type}.",
                "",
                "Lines starting with '#' are ignored.",
            ],
        )

        return editor.edit_apply_loop(
            apply=lambda new: self._apply_target_list(
                source=source,
                relation_type=relation_type,
                targets=new,
                if_match=if_match,
            ),
            apply_exception_class=DebusineError,
            print_exception=self.print_server_error,
            content_description="collection relation",
        )

    @override
    def run(self) -> None:
        if self.args.edit and (
            self.args.append
            or self.args.prepend
            or self.args.remove
            or self.args.set
        ):
            print(
                "--edit cannot be combined with --append, --prepend, "
                "--remove, or --set",
                file=sys.stderr,
            )
            raise SystemExit(1)

        with self._api_call_or_fail():
            current_relations, etag = (
                self.debusine.collection_relation_list_for_edit(
                    self.workspace,
                    source=self.args.source,
                    relation_type=self.args.relation_type,
                )
            )

        current = _collection_identifiers_to_strs(current_relations)
        new = list(current)

        if self.args.edit:
            result_applied = self._edit_and_apply_targets(
                source=self.args.source,
                relation_type=self.args.relation_type,
                initial=new,
                if_match=etag,
            )
            to_print = (
                result_applied.model_dump(mode="json")
                if result_applied is not None
                else current
            )
            self._print_yaml(to_print)
            return

        if self.args.set:
            new = self.args.set
        if self.args.append:
            new = new + self.args.append
        if self.args.prepend:
            new = self.args.prepend + new
        if self.args.remove:
            new = [target for target in new if target not in self.args.remove]

        if new != current:
            try:
                result_applied = self._apply_target_list(
                    source=self.args.source,
                    relation_type=self.args.relation_type,
                    targets=new,
                    if_match=etag,
                )
            except DebusineError as error:
                print("Error setting relation targets:\n", file=sys.stderr)
                self.print_server_error(error)
                raise SystemExit(3)

            self._print_yaml(result_applied.model_dump(mode="json"))



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Common infrastructure for testing the debusine Cli commands."""

import argparse
import logging
from collections.abc import Callable
from typing import Any
from unittest import mock
from unittest.mock import MagicMock

from debusine.client.cli import Cli
from debusine.client.commands.base import Command, DebusineCommand
from debusine.client.debusine import Debusine
from debusine.client.tests import base


class BaseCliTests(base.BaseCliTests):
    """Basic functionality to implement tests for the Cli class."""

    def build_parsed_namespace(self, **kwargs: Any) -> argparse.Namespace:
        """
        Build a DebusineCommand for tests.

        :param kwargs: key-value pairs to set in the mock argparse namespace
          passed to the DebusineCommand
        """
        kwargs.setdefault("server", None)
        kwargs.setdefault("scope", None)
        kwargs.setdefault("silent", False)
        kwargs.setdefault("debug", False)
        kwargs.setdefault("yaml", False)
        if "config_file" not in kwargs:  # pragma: no cover
            kwargs["config_file"] = self.create_config_file().as_posix()

        args = argparse.Namespace()
        for k, v in kwargs.items():
            setattr(args, k, v)

        return args

    def command[C: Command](self, command_class: type[C], **kwargs: Any) -> C:
        """Create an argumentless constructor function for a command class."""
        parser = Cli.build_argument_parser()
        namespace = self.build_parsed_namespace(**kwargs)
        return command_class(parser, namespace)

    def patch_build_debusine_object(self) -> MagicMock:
        """
        Patch _build_debusine_object. Return mock.

        The mocked object return_value is a MagicMock(spec=Debusine).

        Keyword arguments correspond to methods of the resulting Debusine
        object to mock.
        """
        patcher = mock.patch.object(
            DebusineCommand, "_build_debusine_object", autospec=True
        )
        mocked = patcher.start()
        mock_debusine = mock.create_autospec(spec=Debusine)
        mock_debusine._logger = logging.getLogger("debusine.client.tests")
        mocked.return_value = mock_debusine
        self.addCleanup(patcher.stop)

        return mocked

    def assertShowsError(self, func: Callable[[], None]) -> Exception:
        """
        Call the function and ensures it calls show_error.

        :returns: the exception passed to show_error.
        """
        with mock.patch(
            "debusine.client.commands.base.Command.show_error",
            side_effect=SystemExit(42),
        ) as show_error:
            stderr, stdout = self.capture_output(
                func, assert_system_exit_code=42
            )
        show_error.assert_called_once()
        self.assertEqual(stderr, "")
        self.assertEqual(stdout, "")
        exception = show_error.call_args[0][0]
        assert isinstance(exception, Exception)
        return exception



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Test utility classes to help to test the Worker."""

import datetime as dt
from collections.abc import Callable
from dataclasses import dataclass, field
from enum import Enum, auto
from typing import Any, override

from aiohttp import web
from aiohttp.test_utils import AioHTTPTestCase
from aiohttp.web_exceptions import HTTPCreated, HTTPNoContent
from multidict import MultiMapping

from debusine.test.test_utils import create_work_request_response


@dataclass
class ServerConfig:
    """Settings that define server's behaviour when receiving requests."""

    registered_token: str | None = None

    # HTTP status code returned by /api/1.0/worker/register/ endpoint
    register_response_status_code: int = HTTPCreated.status_code

    # HTTP status code returned by /api/1.0/worker/dynamic-metadata/
    dynamic_metadata_status_code: int = HTTPNoContent.status_code

    # The response body when a websocket connection is established
    # If None the server only sends a "connected" message
    # (/api/ws/1.0/worker/connect/)
    connect_response_bodies: list[dict[str, str]] = field(default_factory=list)

    class GetNextForWorkerResponse(Enum):
        # Next response for GET /api/1.0/work-request/get-next-for-worker/
        VALID_WORK_REQUEST = auto()  # well formatted WorkRequest
        NO_WORK_REQUEST_PENDING = auto()  # HTTP 204 empty content
        INVALID_WORK_REQUEST = auto()  # missing mandatory fields
        INVALID_JSON = auto()  # server did not return valid JSON

    get_next_for_worker: GetNextForWorkerResponse | web.Response = (
        GetNextForWorkerResponse.VALID_WORK_REQUEST
    )

    RESPONSES = {
        'connected': {'text': 'connected'},
        'request_dynamic_metadata': {'text': 'request_dynamic_metadata'},
        'work_request_available': {'text': 'work_request_available'},
        'token_disabled_error': {
            'reason': 'The token is disabled',
            'reason_code': 'TOKEN_DISABLED',
        },
    }


@dataclass
class RequestStorage:
    """Storage for the latest requests received by the server's endpoints."""

    method: str | None = None
    path: str | None = None
    json_content: dict[Any, Any] | None = None

    # the token received from the Worker matched the registered_token
    # from the server's registered_token
    is_authenticated: bool = False

    # received headers
    headers: MultiMapping[str] | None = None


class DebusineAioHTTPTestCase(AioHTTPTestCase):
    """
    Implement get_application(): returns a Debusine test server.

    Tests can set up the test server behaviour (see
    DebusineAioHTTPTestCase.server_config) and saves the latest request in
    DebusineAioHTTPTestCase.server_latest_request.
    """

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        """Initialize configuration of the server."""
        super().__init__(*args, **kwargs)

        # server_config changes what the server returns to the Worker:
        # errors, responses, requests, etc.
        self.server_config = ServerConfig()

        # server_latest_requests stores the requests. Tests use it
        # to assert that request were done correctly
        self.server_requests: list[RequestStorage] = []

    @override
    def setUp(self) -> None:
        super().setUp()
        self.work_request = create_work_request_response(
            id=52,
            task_type="Worker",
            task_name="sbuild",
            created_at=dt.datetime.fromisoformat("2022-01-05T11:14:22.242178Z"),
            started_at=None,
            completed_at=None,
            duration=None,
            worker=58,
            task_data={
                "input": {
                    "source_artifact": 5,
                },
                "environment": "debian/match:codename=bullseye",
                "build_architecture": "amd64",
                "build_components": [
                    "any",
                    "all",
                ],
            },
            dynamic_task_data=None,
            priority_base=0,
            priority_adjustment=0,
            status="running",
            result="",
            artifacts=[],
            workspace="TestWorkspace",
        )

    @property
    def server_latest_request(self) -> RequestStorage | None:
        """Return the latest request done to the server or None."""
        if self.server_requests:
            return self.server_requests[-1]
        else:
            return None

    async def _update_latest_request_information(
        self, request: web.Request
    ) -> None:
        """Save request information to self.server_latest_request."""
        request_storage = RequestStorage()
        request_storage.method = request.method
        request_storage.path = request.path

        if request.method in ('PUT', 'POST'):
            request_storage.json_content = await request.json()

        request_storage.is_authenticated = (
            'token' in request.headers
            and request.headers['token'] == self.server_config.registered_token
        )

        request_storage.headers = request.headers

        self.server_requests.append(request_storage)

    @override
    async def get_application(self) -> web.Application:
        """
        Return an instance of aiohttp.web.Application with debusine endpoints.

        Return server that will save requests for the tests to verify
        what was requested. Allows (via self.server_config) to change its
        behaviour.
        """

        async def register(request: web.Request) -> web.Response:
            await self._update_latest_request_information(request)

            status = self.server_config.register_response_status_code

            body = ''
            if status != HTTPCreated.status_code:
                body = 'error'

            return web.Response(body=body, status=status)

        async def dynamic_metadata_received(
            request: web.Request,
        ) -> web.Response:
            await self._update_latest_request_information(request)

            return web.Response(
                status=self.server_config.dynamic_metadata_status_code
            )

        async def connect(request: web.Request) -> web.WebSocketResponse:
            ws = web.WebSocketResponse()
            await ws.prepare(request)

            await collect_request_information(request)

            for body in self.server_config.connect_response_bodies or [
                self.server_config.RESPONSES["connected"]
            ]:
                await ws.send_json(body)

            return ws

        def _create_valid_work_request() -> web.Response:
            return web.json_response(self.work_request.model_dump(mode="json"))

        async def get_next_for_worker(request: web.Request) -> web.Response:
            """Return next_work_request."""
            await self._update_latest_request_information(request)

            res = ServerConfig.GetNextForWorkerResponse

            responses: dict[
                ServerConfig.GetNextForWorkerResponse,
                Callable[[], web.Response],
            ] = {
                res.VALID_WORK_REQUEST: lambda: _create_valid_work_request(),
                res.NO_WORK_REQUEST_PENDING: lambda: web.json_response(
                    status=HTTPNoContent.status_code
                ),
                res.INVALID_WORK_REQUEST: lambda: web.json_response(
                    {'id': 'id should be an int'}
                ),
                res.INVALID_JSON: lambda: web.Response(
                    body='something that is not JSON'
                ),
            }

            if isinstance(self.server_config.get_next_for_worker, web.Response):
                return self.server_config.get_next_for_worker
            else:
                return responses[self.server_config.get_next_for_worker]()

        async def collect_request_information(
            request: web.Request,
        ) -> web.Response:
            """Save the request to be tested."""
            await self._update_latest_request_information(request)
            return web.Response(status=HTTPNoContent.status_code)

        debusine = web.Application()
        debusine.router.add_post('/api/1.0/worker/register/', register)
        debusine.router.add_get('/api/ws/1.0/worker/connect/', connect)
        debusine.router.add_put(
            '/api/1.0/worker/dynamic-metadata/',
            dynamic_metadata_received,
        )

        debusine.router.add_post(
            '/api/1.0/work-request/get-next-for-worker/', get_next_for_worker
        )

        debusine.router.add_get(
            '/collect_request_information/', collect_request_information
        )
        debusine.router.add_put(
            '/collect_request_information/', collect_request_information
        )
        debusine.router.add_post(
            '/collect_request_information/', collect_request_information
        )
        return debusine
